/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770   */
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566   */
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/
/*                                                                              */
/* linux-driver.c - Hytec driver for Linux Modules                              */
/*                                                                              */
/* (C)2006 Hytec Electronics Ltd.                                               */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Revision history: (comment and initial revisions)                           */
/*                                                                              */
/*  vers.       revised         modified by                                     */
/*  -----       -----------     -------------------------                       */
/*  1.1			29-SEP-2005		D. Nineham, Hytec Electronics Ltd.				*/
/*                              Initial version.                                */
/*                                                                              */
/*  1.2			05-SEP-2006		D. Nineham, Hytec Electronics Ltd.				*/
/*                              1. Delay Loops replaced.                        */
/*                                                                              */
/*  1.3			09-OCT-2006		D. Nineham, Hytec Electronics Ltd.				*/
/*                              1. Support for 5331 Added.                      */
/*                                                                              */
/*  1.4			20-SEP-2007		D. Nineham, Hytec Electronics Ltd.				*/
/*                              1.                                              */
/*                                                                              */
/*  1.5			08-JUL-2008		D. Nineham / M. Woodward, Hytec Electronics Ltd.*/
/*                              1. Added 32bit read / write IOCTL functions     */
/*                                                                              */
/*  1.6			11-NOV-2008		M. Woodward, Hytec Electronics Ltd.				*/
/*                              1. Added support for PCI 5332 card				*/
/*                                                                              */
/*  1.7			14-JAN-2008		M. Woodward, Hytec Electronics Ltd.				*/
/*                              1. Fixed problem with 16bit memory writes		*/
/*                                 overwriting adjacent mem locations			*/
/*																				*/
/*  1.8			26-MAR-2009		J. Chen, Hytec Electronics Ltd.					*/
/*                              1. Fixed driver reload problem (region not 		*/
/*                                 released up on close)						*/
/*								2. Added interrupt support for 5331/3331		*/
/*										THE INTERRUPT MECHANISM					*/
/*								3. Added release function up on object destroy.	*/
/*								   This is used for unregistering interrupt		*/
/*								   handlers for those who want interrupt 		*/
/*								   support.										*/
/*								4. Added fasync function for process 			*/
/*								   registration who interested in interrupt		*/
/*								5. Modified read function to return interrupt	*/
/*								   vector to the user. So this read doesn't		*/
/*								   do anything else								*/
/*								6. Sorted out the interrupt support for 9010	*/
/*								7. Added support for Hytec 6335 PCI Express 	*/
/*								   card											*/
/*  1.9			19-MAY-2009		J. Chen, Hytec Electronics Ltd.					*/
/*								1. Changed interrupt level fetch by using 		*/
/*								   pci_get_dev call (old fashion but works)		*/
/*								2. Changed removing resource only when memory 	*/
/*								   mapping and irq assigning are successful		*/
/*  					        3. Commented out kprint in open and release		*/
/*  2.0			02-JUN-2009		J. Chen, Hytec Electronics Ltd.					*/
/*  					        Use a macro define to determine kprint     		*/
/*  3.0			17-SEP-2009		J. Chen, Hytec Electronics Ltd.					*/
/*  					        Added memory mapping from physical address to 	*/
/*  					        the user space 	                                */
/*  3.1			28-SEP-2009		J. Chen, Hytec Electronics Ltd.					*/
/*  					        Added fast interrupt call back to the user      */
/*  					        space via block read function of the driver     */
/*  3.2			19-JAN-2010		J. Chen, Hytec Electronics Ltd.					*/
/*  					        Added one ioctl call for returning "Base ADD3"  */
/*  					        for user space memory mapping                   */
/*  3.3			14-JUL-2010		J. Chen, Hytec Electronics Ltd.					*/
/*  					        Minor change for version control  				*/
/*  3.4			15-JUL-2010		J. Chen, Hytec Electronics Ltd.					*/
/*  					        Commented out the the version detection			*/
/*  3.5			15-JUL-2010		J. Chen, Hytec Electronics Ltd.					*/
/*  					        Added support for Hytec uTCA 7002 card			*/
/*  3.6			11-OCT-2010		J. Chen, Hytec Electronics Ltd.					*/
/*  					        Added interrupt support for Hytec uTCA 7002 card*/
/*  3.7			11-OCT-2010		K. Jogia, Hytec Electronics Ltd.				*/
/*  					        Added BAR 2 & 1 in IOCTL, & Disable interrupts	*/
/*  4.0			19-NOV-2010		J. Chen, K. Jogia, Hytec Electronics Ltd.		*/
/*  					        Converted the driver for supporting multiple 	*/
/*								devices											*/
/*  4.1			26-NOV-2010		J. Chen, Hytec Electronics Ltd.					*/
/*  					        Changed Read function to return both interrupt 	*/
/*								vector and carrier id for multiple carrier case.*/
/*  4.2			30-NOV-2010		J. Chen, Hytec Electronics Ltd.					*/
/*  					        Modified read function to handle multiple 		*/
/*								carrier interrupts.								*/
/*  4.3			01-DEC-2010		J. Chen, Hytec Electronics Ltd.					*/
/*  					        Modified the interrupt service routine to pass	*/
/*								some spurious(?) (IOC9010) interrupts over.		*/
/*								Apperently this is caused by the error interrupt*/
/*								mask. Now these mask bits have been cleared. 	*/
/*  4.4			06-DEC-2010		J. Chen, Hytec Electronics Ltd.					*/
/*  					        Added clean up on error for device registration.*/
/*								Force the driver to use dynamic numbering to	*/
/*								complaint to the main stream.					*/
/*  4.5			14-FEB-2011		J. Chen, Hytec Electronics Ltd.					*/
/*								Changed error wording in io/mem region requests */
/*								to indicate other driver is also uses them		*/
/*																				*/
/*								Modified interrupt service routine to return	*/
/*								IRQ_NONE if the interrupt is not for this 		*/
/*								driver. But if it is and not interrupt source	*/
/*								found, it is a spurious interrupt then return	*/
/*								IRQ_HANDLED to kill it							*/
/*																				*/
/*								Added Hytec 8515/8516 card detection to check 	*/
/*								whether these devices have been used by UART	*/
/*								driver. It they are, the interrupt service		*/
/*								routine in this driver wouldn't handle serail 	*/
/*								related interrupts and leave them to the UART.	*/
/*  5.0			14-MAR-2011		J. Chen, Hytec Electronics Ltd.					*/
/*								Major changes for handling interrupts. Now the  */
/*								driver can handle multiple devices. Each device	*/
/*								can have multiple files. And each file can have */
/*								multiple IP cards (vectors). The ISR can 		*/
/*								dispatch the interrupt vector to the correct	*/
/*								IP card user end callback function.				*/
/*  5.1			29-NOV-2011		J. Chen, Hytec Electronics Ltd.					*/
/*								Modified to suit the interrupt mechanism for 	*/
/*								new kernel starting from 2.6.23. Yet still		*/
/*								maintains the backwards compatibility			*/
/*  5.2			06-SEP-2012		J. Chen, Hytec Electronics Ltd.					*/
/*								Fixed 6335 IP ID detection error			 	*/
/*								Added 7003 uTCA support						 	*/
/*  5.3			24-FEB-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Fixed wrong report for IP module of empty slot	*/
/*  5.4			17-APR-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Removed BAR 1 request for 7003					*/
/*  5.5			22-APR-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Corrected typos									*/
/*  5.6			28-APR-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Fixed ioctl call for higher kernel above 		*/
/*								2.6.36											*/
/*  5.7			07-MAY-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Changed carrier card registers access for 7003	*/
/*								to memory mapping instead of IO port mapping.	*/
/*  5.8			30-JUN-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Modified carrier access for 7003 to memory		*/
/*								access.											*/
/*  5.9			10-JUL-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Fixed carrier access for 7003 to memory			*/
/*								access bug.										*/
/*  6.0			21-JUL-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Fixed carrier card interrupt issue for 7003		*/
/*  6.1			07-AUG-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Modified the driver to suit higher version of	*/
/*								Linux (>2.6.43 from Fedora) which cannot access	*/
/*								I/O port space by inw/outw, but inb/outb works. */ 
/*								Note, the version boundary could be earlier 	*/
/*								than 2.6.43.									*/
/*  6.2			12-AUG-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Fixed interrupt control issue on 6335 carrier 	*/
/*								card for higher version of Linux.				*/
/*  6.3			13-AUG-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Got ride of uint32_t define problem 			*/
/*  6.4			11-SEP-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Added 6335 PCIe carrier model and version read	*/
/*  6.5			02-OCT-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Fixed a bug for IOC9010 carrier registers access*/
/*  6.6			28-NOV-2014		J. Chen, Hytec Electronics Ltd.					*/
/*								Fixed a bug for 7003 carrier registers access	*/
/*								This now uses memory mapping, not I/O mapping	*/
/*																				*/
/********************************************************************************/


/*******************************************************************************
*
* COMPILATION OPTIONS
*
*   __DEBUG__
*       Define this if you want diagnostics output from the driver as it runs.
*       While very useful for debugging (hence the name), this takes up a lot of
*       CPU when running live.
*
*   __DIRECT_IO__
*       Define this if you want to include the direct I/O routines in the
*       driver. These routines access the FIFO directly and are no longer for
*       general use.
*
*******************************************************************************/

#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/time.h>
#include <linux/cdev.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/wait.h>
#include <linux/spinlock.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/version.h>
#include <asm/uaccess.h>
#include "9010LinuxDriver.h"
#include "linux/irq.h"

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10))
#include <linux/config.h>
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
#include <linux/autoconf.h>
#else
#include <generated/autoconf.h>
#endif
#endif


MODULE_AUTHOR("Darrell Nineham, Jim Chen, Kishan Jogia, Mark Woodward");
MODULE_DESCRIPTION("Hytec 9010/5331/5332/6335/7002 linux driver");

/*******************Hytec module specific ***********************/
#define BASE_ADDR	0x300	/* Default register base address    */
#define CRATE_NO	0		/* Default crate number 	        */
#define FALSE		(1==2)
#define OFF_LS		0		/* Position of ls 16 bits in 32     */
#define OFF_MS		1		/* Position of ms 16 bits in 32     */
#define TRUE		(1==1)
#define VME_BERR	0x80	/* Mask for testing Bus error	    */
#define VME_EAI 	0x20	/* Enable address increment	        */

#define DLOW        0x0     /* Data      low         register   */
#define DHIGH		0x2		/* Data      high	     register   */
#define ALOW		0x4		/* Address   low	     register   */
#define AHIGH		0x6		/* Address   high	     register   */
#define AM		    0x8		/* Address   modifier	 register   */
#define IACK		0xa		/* Interrupt acknowledge register   */
#define CSR		    0xc		/* Control   and Status  register   */

/* ------------------------------------------------------------------------ */
/* The following Macros are intended to reduce the possibility of error in  */
/* the written code and to make that code more readable.		    */
/* ------------------------------------------------------------------------ */

#define address_16 outw(vme_addr, ioc9010dev->CarrierAddr+ALOW) /* Set up short address */
#define address_24 outw(vme_addr, ioc9010dev->CarrierAddr+ALOW); \
		           outw((vme_addr>>16), ioc9010dev->CarrierAddr+AHIGH)
#define address_32 outw(vme_addr, ioc9010dev->CarrierAddr+ALOW); \
		           outw((vme_addr>>16), ioc9010dev->CarrierAddr+AHIGH);

#define am_short   outw(0x29, ioc9010dev->CarrierAddr + AM)
#define am_short32 outw(0xa9, ioc9010dev->CarrierAddr + AM)
#define am_stand   outw(0x39, ioc9010dev->CarrierAddr + AM)
#define am_stand32 outw(0xb9, ioc9010dev->CarrierAddr + AM)
#define am_long    outw(0x09, ioc9010dev->CarrierAddr + AM)
#define am_long32  outw(0x89, ioc9010dev->CarrierAddr + AM)

#define dis_inc    outw((inw(ioc9010dev->CarrierAddr+CSR) & 0x1f), ioc9010dev->CarrierAddr+CSR)
#define enb_inc    outw((inw(ioc9010dev->CarrierAddr+CSR) & 0x1f) + VME_EAI, ioc9010dev->CarrierAddr+CSR)

/*
#define read_8	   *(char *)pdata=inp(base_addr+DLOW)// Read 8	bits of data
*/
#define read_16    *(short *)pdata=inw(ioc9010dev->CarrierAddr+DLOW) /* Read 16 bits data */

#define read_32    *(short *)((short *)pdata+OFF_LS)=inw(ioc9010dev->CarrierAddr + DLOW); \
		   *(short *)((short *)pdata+OFF_MS)=inw(ioc9010dev->CarrierAddr + DHIGH)
/*
#define bread_8    *pdata++ = inp (base_addr+DLOW)

*/

#define bread_16   *pdata++ = inw(ioc9010dev->CarrierAddr+DLOW)

#define bread_32   *pdata = inl(ioc9010dev->CarrierAddr+DLOW)

/*
#define bread_32   *(short *)((short *)pdata+OFF_LS)=inpw(base_addr + DLOW); \
		   *(short *)((short *)pdata+OFF_MS)=inpw(base_addr + DHIGH); \
                   pdata++
					// Return TRUE if no bus error
*/

#define return_be return((inw(ioc9010dev->CarrierAddr + CSR) & VME_BERR) == 0)

#define write_8    outw(data, ioc9010dev->CarrierAddr+DLOW)   /* write  8 bits of data */
#define write_16   outw(data, ioc9010dev->CarrierAddr+DLOW)   /* write 16 bits of data */

/*
#define write_32   pdatw = (short *)&data; \
		   outpw(base_addr + DHIGH,*(pdatw + OFF_MS)); \
		   outpw(base_addr + DLOW ,*(pdatw + OFF_LS))
#define bwrite_8   outp (base_addr + DLOW, *pdata++)
*/

#define bwrite_16  outw(*pdata++, ioc9010dev->CarrierAddr + DLOW)


#define bwrite_32  outw(*(short *)((short *)pdata+OFF_MS),ioc9010dev->CarrierAddr + DHIGH);\
		   outw(*(short *)((short *)pdata+OFF_LS),ioc9010dev->CarrierAddr + DLOW);\
		   pdata++
/*
#define bwrite_32  outpw(base_addr + DHIGH,*(short *)((short *)pdata+OFF_MS));\
		   outpw(base_addr + DLOW ,*(short *)((short *)pdata+OFF_LS));\
		   pdata++
*/


/******* Supported Hytec PCI Devices ********/
static struct pci_device_id ids[] = {
	{ PCI_DEVICE(0x124B, 0x0040), },
	{ PCI_DEVICE(0x1196, 0x9010), },
	{ PCI_DEVICE(0x1196, 0x5331), },
	{ PCI_DEVICE(0x1196, 0x5332), },
	{ PCI_DEVICE(0x1196, 0x6335), },
	{ PCI_DEVICE(0x1196, 0x7002), },
	{ PCI_DEVICE(0x1196, 0x7003), },
	{ 0, } 
};
MODULE_DEVICE_TABLE(pci, ids);


/************** Gloabal Variables ***********/

#define MAXDEVNUM				21			/* maximum device number   */
#define MAXFILENUM				64			/* maximum file number on one device  */
#define	MAXIPNUM				6			/* maximum IP numbers on a single carrier, IOC9010 */

/*int ioc_major = 60;*/						/* release on 02 June 2009 */
int 						ioc_major = 0;	/* Force it to use dynamic numbering */
int 						ioc_minor = 0;
dev_t 						dtMajorMinor;
module_param(ioc_major, int, 0);

char 						*buffer;        /* Begin of Kernel buffer */

/*****Interrupt operation data area **********/
/* 	  added by Jim, 26-03-2009               */
#define PLXLocalConfigCSR		0x4c  		/* Interrupt Control register for PLX 9030, used by IOC9010 */
#define Hytec7002CSR			0x00		/* 7002 Control register   */
#define PLXLocalConfigCSR8311	0x68  		/* Interrupt Control register for PLX PEX 8311, used by 6335 */

/* The following structure is for interrupt dispatching for a single 
 * file created by the user on a device. One PCI/PCIe device can 
 * handle maximum of 64 files defined by MAXFILENUM in this driver. 
 * This structure records the opening file handler (pointer), the 
 * interrupt vectors of that file and the queuing mechanism stuffs in 
 * order for the device interrupt service routine to dispatch the 
 * correct vector to the IP card service routine to the user end. The 
 * mechanism is based on the following facts.
 *
 *   -- One Linux operating system device driver can create multiple 
 *      devices in /dev directory
 *
 *   -- A Linux operating system device can be opened multiple times by
 *      multiple files. These files can be in the same process or in 
 *      different processes. Yet there is only one piece of code 
 *      associating to all these files of that device in the kernel space.
 *
 *   -- A single file can be used by one IP card or shared by multiple 
 *      IP cards hence one single file could have one to many vectors or 
 *      interrupt requests. The maximum case is the IOC9010 which could
 *      have 6 IPs that use one single file so that we define an array 
 *      of 6 for vector.
 *
 *   -- A device can have only one interrupt service routine no matter 
 *      how many files are associated to it. So this isr is shared by
 *      all the open files of that device. This is also why only the 
 *      device structure is passed to the isr.
 *
 * Interrupt implementation mechanism
 * =====================================
 *
 *   -- When device driver is loaded, it identifies one to many PCI/PCIe 
 *      devices and creates one to many IOC9010_dev device structures.
 *
 *   -- Each device structure has 64 this irq_dispatch strucutres.
 *
 *   -- The loading process also creates one to many devices in the
 *      /dev directory
 *
 *   -- When user opens a file by calling IOC9010_open on a device, the 
 *      driver registers its file pointer in this structure.
 *
 *   -- When user sets (connects) up an IP card interrupt, the driver 
 *      registers its vector to the proper file structure by calling 
 *      OP_REGISTER_VECTOR ioctl call. The connect function also creates
 *      a thread which allows the IP card(s) to wait for interrupts by 
 * 	    calling IOC9010_read block function which in turn waits on the 
 *      proper file queue. 
 *
 *      NOTE 1: even though only one piece of kernel code for one 
 *      device, the IOC9010_read function can be re-entried by multiple 
 *      files because each file is literally waiting in the system queue. 
 *
 *      NOTE 2: the thread created by the connect function needs to be 
 *      created only once for all devices that use the same file. The user
 *      API (for EPICS applications, it is the drvHyLinuxCarrier.c file
 *      and for Linux Hytec API application, it is the IOCGeneric.c file) 
 *      creates a mapping array that maps vectors to callback functions.
 *
 *   -- When an interrupt has been generated by one of the IP cards 
 *      registered to this device, the isr loops all files and vectors
 *      to identify which file (IP) generates it. Then it wakes up the 
 *      proper queue waiting in the IOC9010_read function to service the 
 *      correct user (the API). Then the API calls the proper callback 
 *      of the user code by the vector-callback mapping.
 */
struct irq_dispatch{
	struct file 			*filep;			/* file pointer to the associated device */
	atomic_t 				atomCounter; 	/* atomic counter */
	wait_queue_head_t 		queue;			/* wake up queue */
	char 					vector[6];		/* vectors for each open file. Maximum IP cards is on an IOC blade which has 6 hence an array of 6 */
	int						pending;		/* kernel space semaphore */
	char 					buf[10];		/* local buffer for copying vector to the user end */
};

/*****End of interrupt operation data area **********/

struct IOC9010_dev {
	struct IOC9010_dev* 	next;				/* store next device pointer */
	int	devnum;									/* deice number */
	struct irq_dispatch 	irqdp[MAXFILENUM];
	
	void 					*VirtureMemBaseAddr;
	void 					*VirtureCarrierRegBaseAddr;				/* this is for 7003 only since the change over from IO port to memory */
	unsigned int 			CarrierAddr;  		/* Base Address of Carrier Card Registers - I/O Region */
	unsigned char 			CarrierIRQ;   		/* Interrupt Number of the Carrier Card   - IRQ */
	unsigned long 			IOCMemoryAddr;  	/* IOC memory Base for IP cards - memory Region */
	
	u16 					IPCard[6];          /* Array of Industry Pack Cards Found */
	int 					VENDOR_ID;
	int 					DEVICE_ID;
	
	int 					irqState;			//for interrupt initialisation successful flag
	int 					bar1State;			//for BAR1 IO/memory allocation successful flag
	int 					bar2State;			//for BAR2 IO/memory allocation successful flag
	int 					bar3State;			//for BAR3 IO/memory allocation successful flag

	/*****Interrupt operation data area **********/
	/* 	  added by Jim, 26-03-2009               */
	unsigned int 			PLXLocalConfigBAR;	/* Base Address of PLX local configuration Registers - I/O Region.*/
	unsigned int 			PCIexpressBAR;  	/* Base Address of PCI express configuration Registers (BAR0)- I/O Region.*/
	void 					*addrPCI;
	struct pci_dev 			*ppciDev;			/* store the associated pci device struct pointer */

	/* These part is currently not used. It was for interrupt test but the fast interrupt mechanism is used instead */
	struct fasync_struct 	*async_queue;		/* processes queue for interrupt notification. added by Jim on 26-03-2009 */
	unsigned short 			vector;				/* for interrupt. added by Jim on 26-03-2009 */
	int 					pid[TASK_MAXPID];	/* for interrupt. added by Jim on 26-03-2009 */
    int 					pidNumber;          /* for interrupt. added by Jim on 26-03-2009 */
    int 					signal;             /* signal for notification. default to SIGIO. for interrupt. added by Jim on 26-03-2009 */

  	spinlock_t 				device_lock;   		/* access to device registers */
	unsigned long 			lock_flags;
	struct cdev 			cdev;	  			/* Char device structure */
};

struct IOC9010_dev* 		devlist=NULL;		/* the pointer to the devices */

#define dev_lock  	spin_lock_irqsave(&ioc9010dev->device_lock, ioc9010dev->lock_flags)
#define dev_unlock  spin_unlock_irqrestore(&ioc9010dev->device_lock, ioc9010dev->lock_flags)




/**********************************************************

    NAME:	vme_block_read_short_16(struct IOC9010_dev* ioc9010dev, vme_addr,pdata,length)

    Description:

    Function - Read a block of VME locations defined by a 16
    bit address (vme_addr) using AM 29(hex) and put the words 
	read into the array defined by the word pointer (pdata).
	
	Parameters:

    Inputs:
	       long    vme_addr - VME address to be accessed	    
		   short   *pdata   - Pointer to target 16 bits
		   long	   length   - Number of words to read


    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 11-OCT-2006     DAN       1.0       Intial Version.
 22-NOV-2010     JSC       2.0       Modified to support multiple devices.

***********************************************************/
int	vme_block_read_short_16(struct IOC9010_dev* ioc9010dev, long vme_addr, short *pdata, long length)
{
	long	iloop;

//	dev_lock;			/* spinlock code */
	am_short;			/* Set up the Address Modifier	    */
//	enb_inc;			/* Enable address increment			*/
	for (iloop=0; iloop<length; iloop++)
	{   
		address_16;			/* Set up the 16 bits of address    */
        bread_16;			/* Read 16 bits of data 	    */
		vme_addr+=2;		/* Add 2 to the VME address		*/
	}
//	dis_inc;			/* Disable address increment	    */
//	dev_unlock;			/* spinlock code */
	return_be;			/* Return TRUE if no bus error	    */
}


/**********************************************************

    NAME:	vme_block_write_short_16(struct IOC9010_dev* ioc9010dev, vme_addr,pdata,length)

    Description:

    Function - Write a VME location defined by a 16 bit 
	address (vme_addr) using AM 29(hex) and taking the data 
	from the location defined by the word pointer (data)
	
	Parameters:

    Inputs:
	       long    vme_addr - VME address to be accessed	    
		   short   *pdata   - Pointer to target 16 bits
		   long	   length   - Number of words to read


    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 11-OCT-2006     DAN       1.0       Intial Version.
 22-NOV-2010     JSC       2.0       Modified to support multiple devices.

***********************************************************/
int	vme_block_write_short_16(struct IOC9010_dev* ioc9010dev, long vme_addr, short *pdata, long length)
{
	long	iloop;

//	dev_lock;			/* spinlock code */
	am_short;			/* Set up the Address Modifier	    */
//	enb_inc;			/* Enable address increment			*/
	for (iloop=0; iloop<length; iloop++)
	{
		address_16;			/* Set up the 16 bits of address    */
	   	bwrite_16;			/* Write 16 bits of data	    */
		vme_addr+=2;		/* Add 2 to the VME address		*/
	}
//	dis_inc;			/* Disable address increment	    */
//	dev_unlock;			/* spinlock code */
	return_be;			/* Return TRUE if no bus error	    */
}


/**********************************************************

    NAME:	vme_block_read_long_16(struct IOC9010_dev* ioc9010dev, vme_addr,pdata,length)

    Description:

    Function - Read a block of VME locations defined by a 
	32 bit address (vme_addr) using AM 09(hex) and put the 
	words read into the array defined by the word pointer 
	(pdata).


	Parameters:

    Inputs:
	       long    vme_addr - VME address to be accessed	    
		   short   *pdata   - Pointer to target 16 bits
		   long	   length   - Number of words to read


    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 11-OCT-2006     DAN       1.0       Intial Version.
 22-NOV-2010     JSC       2.0       Modified to support multiple devices.

***********************************************************/
int	vme_block_read_long_16(struct IOC9010_dev* ioc9010dev, long vme_addr, short *pdata, long length)
{
	long	iloop;

//	dev_lock;			/* spinlock code */
	/* address_32;		 Set up the 32 bits of address    */
    outw((vme_addr & 0x0000FFFF), ioc9010dev->CarrierAddr+ALOW);
	outw((vme_addr>>16)         , ioc9010dev->CarrierAddr+AHIGH);
	am_long;			/* Set up the Address Modifier	    */
	enb_inc;			/* Enable address increment	    */
	for (iloop=0; iloop<length; iloop++)
	{   bread_16;			/* Read 16 bits of data 	    */
	}
	dis_inc;			/* Disable address increment	    */
//	dev_unlock;			/* spinlock code */
	return_be;			/* Return TRUE if no bus error	    */
}


/**********************************************************

    NAME:	vme_block_write_long_16(struct IOC9010_dev* ioc9010dev, vme_addr,pdata,length)

    Description:

    Function - Writes a VME locati	unsigned short	Address;
on defined by a 32 bit 
	address (vme_addr) using AM 09(hex) and taking the data
	from the location defined by the word pointer (data).

	Parameters:

    Inputs:
	       long    vme_addr - VME address to be accessed	    
		   short   *pdata   - Pointer to target 16 bits
		   long	   length   - Number of words to read


    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 11-OCT-2006     DAN       1.0       Intial Version.
 22-NOV-2010     JSC       2.0       Modified to support multiple devices.

***********************************************************/
int	vme_block_write_long_16(struct IOC9010_dev* ioc9010dev, long vme_addr,short	*pdata, long length)
{
	long	iloop;

//	dev_lock;			/* spinlock code */
	/* address_32;		 Set up the 32 bits of address    */
    outw((vme_addr & 0x0000FFFF), ioc9010dev->CarrierAddr+ALOW);
	outw((vme_addr>>16)         , ioc9010dev->CarrierAddr+AHIGH);
	am_long;			/* Set up the Address Modifier	    */
	enb_inc;			/* Enable address increment	    */
	for (iloop=0; iloop<length; iloop++)
	{   bwrite_16;			/* Write 16 bits of data	    */
	}
	dis_inc;			/* Disable address increment	    */
//	dev_unlock;			/* spinlock code */
	return_be;			/* Return TRUE if no bus error	    */
}


/**********************************************************

    NAME:	vme_block_read_long_32(struct IOC9010_dev* ioc9010dev, vme_addr,pdata,length)

    Description:

    Function - Read a block of VME locations defined by a 
	32 bit address (vme_addr) using AM 09(hex) and put the 
	words read into the array defined by the word pointer 
	(pdata).


	Parameters:

    Inputs:
	       long    vme_addr - VME address to be accessed	    
		   short   *pdata   - Pointer to target 16 bits
		   long	   length   - Number of words to read


    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 11-OCT-2006     DAN       1.0       Intial Version.
 22-NOV-2010     JSC       2.0       Modified to support multiple devices.

***********************************************************/
int	vme_block_read_long_32(struct IOC9010_dev* ioc9010dev, long vme_addr, long *pdata, long length)
{
	long	iloop;

//	dev_lock;			/* spinlock code */
	address_32;			/* Set up the 32 bits of address   */
	am_short32;			/* Set up the Address Modifier	   */
	for (iloop=0; iloop<length; iloop++)
	{   bread_32;		/* Read 32 bits of data	    */
	}
//	dev_unlock;			/* spinlock code */
	return_be;			/* Return TRUE if no bus error	    */
}

/**********************************************************

    NAME:	vme_block_write_long_32(struct IOC9010_dev* ioc9010dev, vme_addr,pdata,length)

    Description:

    Function - Read a block of VME locations defined by a 
	32 bit address (vme_addr) using AM 09(hex) and put the 
	words read into the array defined by the word pointer 
	(pdata).


	Parameters:

    Inputs:
	       long    vme_addr - VME address to be accessed	    
		   short   *pdata   - Pointer to target 16 bits
		   long	   length   - Number of words to read


    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 11-OCT-2006     DAN       1.0       Intial Version.
 22-NOV-2010     JSC       2.0       Modified to support multiple devices.

***********************************************************/
int	vme_block_write_long_32(struct IOC9010_dev* ioc9010dev, long vme_addr, long *pdata, long length)
{
	long	iloop;

//	dev_lock;			/* spinlock code */
	address_32;			/* Set up the 32 bits of address   */
	am_short32;			/* Set up the Address Modifier	   */
	for (iloop=0; iloop<length; iloop++)
	{   bwrite_32;		/* Write 32 bits of data	    */
	}
//	dev_unlock;			/* spinlock code */
	return_be;			/* Return TRUE if no bus error	    */
}


/***********************************************************/

/************ this part is for IOC9010 blade only **********/
#define LCD_LINE_1		0
#define LCD_LINE_2		64
#define LCD_CENTRE_ON	0x8000
#define LCD_LINE_LEN	40


/**********************************************************

    NAME:			LCD_Ready()

    Description:

    Function -		Used to check when the LCD is ready to
					recieve the next command / character.

                    Unfortunantly has to be just delay as
                    Xilinx firmware does not allow LCD Data
                    to be read at present time.

    Parameters:

    Inputs:			Nothing.
    Outputs:		Nothing.
    Return:			Nothing.

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen


**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 23-JAN-2006     DAN     1.1         Intial Version.
 05-SEP-2006     DAN     1.2         Switched away from loops.

***********************************************************/
void LCD_Ready(void)
{
	/* Presently Just Delay - Eventually check LCD Ready Bit */
	msleep(5);
}


/**********************************************************

    NAME:			Clear_LCD(struct IOC9010_dev* ioc9010dev)


    Description:

    Function -		Clears LCD Display.


	Parameters:

    Inputs:			Nothing.
    Outputs:		Nothing.
    Return:			Nothing.


    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 23-JAN-2006     DAN     1.1         Intial Version.
 22-NOV-2010     JSC     2.0         Modified to support multiple devices.

***********************************************************/
void Clear_LCD(struct IOC9010_dev* ioc9010dev)
{
	/* Wait for LCD to be Ready */
	LCD_Ready();

	/* Send Clear LCD Display */
	outw(0x0001, (int)(ioc9010dev->CarrierAddr + REG_LCD_CONTROL));
}


/**********************************************************

    NAME:			Write_LCD(struct IOC9010_dev* ioc9010dev, u16 StartPos,  char* cStringToWrite)

    Description:

    Function -		Writes the passed string at the passed	
					position on the LCD.


    Parameters:

    Inputs:			u16 StartPos		 - Where to write.
					char* cStringToWrite - Pointer to string. 
    Outputs:		Nothing.
    Return:			Nothing.

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen
**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 23-JAN-2006     DAN     1.1         Intial Version.
 22-NOV-2010     JSC     2.0         Modified to support multiple devices.

***********************************************************/
void Write_LCD(struct IOC9010_dev* ioc9010dev, u16 StartPos,  char* cStringToWrite )
{
	LCD_Ready();

	/* Set Start Position on LCD Display */
	/* Centre Message on Top Line LCD */ 
	if 	(StartPos == (LCD_CENTRE_ON | LCD_LINE_1))
	{
		outw((0x0080 | (LCD_LINE_1 + LCD_LINE_LEN/2 - strlen(cStringToWrite)/2)), ioc9010dev->CarrierAddr + REG_LCD_CONTROL);
	}
	/* Centre Message on Bottom Line LCD */ 
	else if (StartPos == (LCD_CENTRE_ON | LCD_LINE_2))
	{
		outw((0x0080 | (LCD_LINE_2 + LCD_LINE_LEN/2 - strlen(cStringToWrite)/2)), ioc9010dev->CarrierAddr + REG_LCD_CONTROL);
	}
	/* Default - At Passed Position */
	else
	{
		outw((0x0080 | StartPos), ioc9010dev->CarrierAddr + REG_LCD_CONTROL);
	}

	LCD_Ready();

	/* Write String to Display */
	while (*cStringToWrite)
	{
		outw(*(cStringToWrite++), ioc9010dev->CarrierAddr + REG_LCD_DATA);
		LCD_Ready();
	}
}


/**********************************************************

    NAME:          InterruptHandler( )

    Description:

    Function - 

    Parameters:

    Inputs:
    Outputs:
    Return:

    Notes:

    Programmer(s):  Jim Chen
					Darrell Nineham

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 23-JAN-2006     DAN     1.1         Intial Version.

 26-MAR-2009     JSC     2.0         Full implementation of interrupt support for IOC9010, 5331/5332 and 6335.
 29-NOV-2011     JSC     2.1         Support new kernel interrupt mechanism.

***********************************************************/
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23))
	irqreturn_t InterruptHandler(int irq, void *dev_id, struct pt_regs *regs)
#else
	irqreturn_t InterruptHandler(int irq, void *dev_id)
#endif
{
	int INTSource, i=0, j, done=0, INTFound;
	unsigned short  sdata;

	struct IOC9010_dev *ioc9010dev = (struct IOC9010_dev *)dev_id;

//   	printk(KERN_NOTICE "%x Test - Interruptted! \n", ioc9010dev->DEVICE_ID);

	dev_lock;			/* spinlock code */
 	    
    switch (ioc9010dev->DEVICE_ID)
    {
    case 0x9010:
    case 0x6335:
    case 0x7002:
    case 0x7003:
    	//get interrupt source
		if(ioc9010dev->DEVICE_ID == 0x7003)
			INTSource = ioread16(ioc9010dev->VirtureCarrierRegBaseAddr + REG_IP_INT);
		else
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))		//From Fedora 15. May need earlier version here?
			INTSource = inw(ioc9010dev->CarrierAddr + REG_IP_INT);
#else
			INTSource = inb(ioc9010dev->CarrierAddr + REG_IP_INT);
#endif

		if(!INTSource) break;

        INTFound = 0;
        for(i=0; i<6; i++)
        {
            if (INTSource & (1 << i*2)) 
            {
                INTFound = 1;
                break;
            }
        }

		if(INTFound)
		{
			/*
			 * When Hytec UART driver is loaded at the same time with this one, 
			 * the UART driver should take care of the serial interrupts hence 
			 * this check. If true, quit here and let the UART do 
			 * the handling. Added on 21/02/2011. JSC
			 */
			if((ioc9010dev->IPCard[i] == 0x8515) || (ioc9010dev->IPCard[i] == 0x8516))
			{
				/* read timer controller register higher 4 bits of the 8515/8516 IP */
    	       	sdata = ioread16(ioc9010dev->VirtureMemBaseAddr + IP_IO_BASE_ADDR + i*0x100 + 0x04*2);
				if((sdata & 0xF0) == 0xAA)			
					return IRQ_NONE;			/* Leave this for the UART. */
			}

			/* Now disable the interrupt first */
			switch(ioc9010dev->DEVICE_ID)
			{
			case 0x7003:
				sdata = ioread16(ioc9010dev->VirtureCarrierRegBaseAddr + Hytec7002CSR);
				sdata &= (~0x0001);
				iowrite16(sdata, ioc9010dev->VirtureCarrierRegBaseAddr + Hytec7002CSR);
				break;
			case 0x7002:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))												//From Fedora 15. May need earlier version here?
				sdata = inw(ioc9010dev->CarrierAddr + Hytec7002CSR);
				sdata &= (~0x0001);
				outw(sdata, ioc9010dev->CarrierAddr + Hytec7002CSR);							//clear IRQ in FPGA. added by Jim on 13-10-2010
#else
				sdata = inb(ioc9010dev->CarrierAddr + Hytec7002CSR);
				sdata &= (~0x0001);
				outb(sdata, ioc9010dev->CarrierAddr + Hytec7002CSR);							//clear IRQ in FPGA. added by Jim on 13-10-2010
#endif				
				break;
			case 0x9010:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))												//From Fedora 15. May need earlier version here?
				sdata = inw(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);
				sdata &= (~0x040);
				outw(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);					//clear IRQ on PLX. added by Jim on 26-03-2009
#else
				sdata = inb(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);
				sdata &= (~0x040);
				outb(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);					//clear IRQ on PLX. added by Jim on 26-03-2009
#endif				
				break;
			case 0x6335:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))												//From Fedora 15. May need earlier version here?
				sdata = inw(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR8311);
				sdata &= (~0x0900);
				outw(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR8311);				//clear IRQ on PLX. added by Jim on 26-03-2009
#else
				sdata = inb(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR8311 + 1);
				sdata &= (~0x09);
				outb(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR8311 + 1);		//clear IRQ on PLX. added by Jim on 26-03-2009					
#endif				
				break;
			}

            ioc9010dev->vector = (ioread16(ioc9010dev->VirtureMemBaseAddr + IP_IO_BASE_ADDR + IP_A_INT_BASE_ADDR + i*0x100)) & 0xFF;

	    	printk(KERN_NOTICE "%x Test - Interruptted! Source bit=%d IP=%d Vector=%x\n", 
	    			ioc9010dev->DEVICE_ID, INTSource, i, ioc9010dev->vector);

            /* 
			 * fast interrupt handling. added on 28-09-2009. J.C 
         	 * disable_irq_nosync(irq); not sure we need this?
			 *
			 * Loops all files and their vectors to try to wake 
			 * up the proper queue. added on 14-03-2011. J.C
			 */
			done = 0;
			for(i=0; i<MAXFILENUM; i++)

			{
				if(ioc9010dev->irqdp[i].filep != NULL)
				{
					for(j=0; j<MAXIPNUM; j++)
					{
						if((ioc9010dev->irqdp[i].vector[j]&0xFF) == (ioc9010dev->vector&0xFF))
						{
							atomic_inc(&ioc9010dev->irqdp[i].atomCounter);
        		 			wake_up_interruptible(&ioc9010dev->irqdp[i].queue);
							done = 1;
							break;
						}
					}
					if(done == 1) break;
				}
			}
         	
/*
            switch(ioc9010dev->signal)
            {
                case SIGIO:
     		       if(ioc9010dev->async_queue)
    		          kill_fasync(&ioc9010dev->async_queue, SIGIO, POLL_IN);		//send signal to whoever interested    			
                   break;
                case SIGUSR1:
                case SIGUSR2:
                    for(i=0; i<ioc9010dev->pidNumber; i++)
                		kill_proc(ioc9010dev->pid[i], ioc9010dev->signal, 1);  		//send out user signal 1 or 2 to all subscribes
                    break;
                default:                                                    //shouldn't get here
                    break;
            }
*/
        }else
		{
			dev_unlock;			/* spinlock code */
			return IRQ_NONE;			/* Leave this for other devices. */
		}
       	break;

	case 0x5331:
    case 0x5332:
   		
		/* Disable Interrupts */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))												//From Fedora 15. May need earlier version here?
        outw(0x0003, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);
#else
        outb(0x03, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);   
#endif
		//outw(0x0007, ioc9010dev->CarrierAddr + CSR);

		//printk(KERN_NOTICE "%x Test - Interrupt Handled %x! \n", ioc9010dev->DEVICE_ID, (int) ioc9010dev->async_queue);

        /* Save Interrupt vector */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))												//From Fedora 15. May need earlier version here?
		ioc9010dev->vector = inw(ioc9010dev->CarrierAddr + IACK);	//save vector in the buffer
#else
		ioc9010dev->vector = inb(ioc9010dev->CarrierAddr + IACK);	//save vector in the buffer
#endif

   		//send signal to user
#ifdef PRINT
   		printk(KERN_NOTICE "%x Test - Signal is about to send...vector=%x ip=%d pidNo=%d\n", ioc9010dev->DEVICE_ID, ioc9010dev->vector, i, ioc9010dev->pidNumber);
#endif

		/* Help me, need code to deal with interrupt dispatching here! */

/*        switch(ioc9010dev->signal)
        {
            case SIGIO:
 		       if(ioc9010dev->async_queue)
 		          kill_fasync(&ioc9010dev->async_queue, SIGIO, POLL_IN);		//send signal to whoever interested    			
               break;
            case SIGUSR1:
            case SIGUSR2:
               for(i=0; i<ioc9010dev->pidNumber; i++)
           		   kill_proc(ioc9010dev->pid[i], ioc9010dev->signal, 1);  		//send out user signal 1 or 2 to all subscribes
               break;
            default:                                                    //shouldn't get here
               break;
        }
*/
        break;
    default:
        break;
    }

	dev_unlock;			/* spinlock code */

    /* return IRQ_NONE; */
    return IRQ_HANDLED;
}


/**********************************************************

    NAME:           IOC9010_open( )

    Description:

    Function - 

    Parameters:

    Inputs:

    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 23-JAN-2006     DAN     1.1         Intial Version.

***********************************************************/
int IOC9010_open(struct inode *inode, struct file *filep)
{
	struct IOC9010_dev *dev;
	int i;

	dev = container_of(inode->i_cdev, struct IOC9010_dev, cdev);
	filep->private_data = dev;

	/* register file in the device irq_dispatch structure array
     * and do its queuing inits.
     */
	for(i=0; i<MAXFILENUM; i++)
	{
		if(dev->irqdp[i].filep == NULL)
		{
			dev->irqdp[i].filep = filep;
			atomic_set(&dev->irqdp[i].atomCounter, 0);
			init_waitqueue_head(&dev->irqdp[i].queue);
			break;
		}
	}

/*	printk(KERN_NOTICE "%x Test - open called. uid=%d pid=%d filep=%x i=%d\n", 
			dev->DEVICE_ID, current->uid, current->pid, (int) filep, i);
*/	
	return 0;
}

/**********************************************************

    NAME:           IOC9010_fasync( )

    Description:	This routine registers the processes who are interested in
					the interrupt so that they can get notified once there is interrupt
					generated
    Function - 

    Parameters:

    Inputs:

    Outputs:
    Return:

    Notes:

    Programmer(s):  Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 26-MAR-2009     JSC     1.1         Intial Version.

***********************************************************/
int IOC9010_fasync(int fd, struct file *filep, int mode)
{
	int status=0;
	struct IOC9010_dev *dev = filep->private_data;
	status = fasync_helper(fd, filep, mode, &dev->async_queue);
	//printk(KERN_NOTICE "%x Test - Process registered. %x\n", ioc9010dev->DEVICE_ID, dev->async_queue); //for debugging
	return status;
}


/**********************************************************

    NAME:           IOC9010_release( )

    Description:	This function is used to release Process registration 
					of interrupt interests
    Function - 


    Parameters:

    Inputs:

    Outputs:
    Return:

    Notes:

    Programmer(s):  Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 26-MAR-2009     JSC     1.1         Intial Version.

***********************************************************/
int IOC9010_release(struct inode *inode, struct file *filep)
{
  	struct IOC9010_dev *ioc9010dev = filep->private_data;
	int i, j;

/*	printk(KERN_NOTICE "Test - Release called.\n");*/

	/* detach file and vectors */
	for(i=0; i<MAXFILENUM; i++)
	{
		if(ioc9010dev->irqdp[i].filep == filep)
		{
			ioc9010dev->irqdp[i].filep = NULL;
			for(j=0; j<MAXIPNUM; j++)
				ioc9010dev->irqdp[i].vector[j] = 0;
			break;
		}
	}
    ioc9010dev->pidNumber = 0;                                  //detach processes
	IOC9010_fasync(-1, filep, 0);								//detach processes registerred.
	return 0;
}

/**********************************************************

    NAME:           IOC9010_mmap_open( )

    Description:	This function is used to open the memory mapping instance. 
					It does nothing at the moment.
    Function - 

    Parameters:
    Inputs:
    Outputs:
    Return:
    Notes:

    Programmer(s):  Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 17-Sep-2009     JSC     1.1         Intial Version.

***********************************************************/
void IOC9010_mmap_open(struct vm_area_struct *vma)
{
    return;
}

/**********************************************************

    NAME:           IOC9010_remap_vm_ops( )

    Description:	Memory mapping IO operation structure 

    Function - 

    Parameters:
    Inputs:
    Outputs:
    Return:
    Notes:

    Programmer(s):  Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 17-Sep-2009     JSC     1.1         Intial Version.

***********************************************************/
static struct vm_operations_struct IOC9010_remap_vm_ops = {
    .open = IOC9010_mmap_open,
};

/**********************************************************

    NAME:           IOC9010_mmap( )

    Description:	This function is used to map memory space to user space 
					directly
    Function - 

    Parameters:
    Inputs:
    Outputs:
    Return:
    Notes:

    Programmer(s):  Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 17-Sep-2009     JSC     1.1         Intial Version.

***********************************************************/
static int IOC9010_mmap(struct file *filep, struct vm_area_struct *vma)
{
#ifdef PRINT
  	struct IOC9010_dev *ioc9010dev = filep->private_data;
#endif
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))			//Before Fedora 15. May need earlier version here?
    vma->vm_flags |= VM_IO | VM_RESERVED;					
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(3,15,0))			//Before Fedora 20. May need earlier version here?
    vma->vm_flags |= VM_IO | VM_DONTEXPAND;					
#else
  	vma->vm_flags |= VM_IO | VM_DONTEXPAND | VM_DONTDUMP; 	//for Fedora 20 and above
#endif
    
    vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10))
	/* remap_page_range((a), (b), (c), (d), (e)) */
	if(remap_page_range(vma, vma->vm_start, vma->vm_pgoff << PAGE_SHIFT,
        vma->vm_end - vma->vm_start, vma->vm_page_prot))
#else
	/* remap_pfn_range((a), (b), (c) >> PAGE_SHIFT, (d), (e)) */
	if(remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
        vma->vm_end - vma->vm_start, vma->vm_page_prot))
#endif
    {
#ifdef PRINT
        printk(KERN_NOTICE "%x Memory mapped not ok.\n", ioc9010dev->DEVICE_ID); 
#endif
        return -EAGAIN;
    }
    vma->vm_ops = &IOC9010_remap_vm_ops;
    IOC9010_mmap_open(vma);                         /* does nothing at the moment */
#ifdef PRINT
    printk(KERN_NOTICE "%x Memory mapped OK -- Start=%x - End=%x\n", ioc9010dev->DEVICE_ID, (int) vma->vm_start, (int) vma->vm_end); 
#endif
	return 0;
}


/**********************************************************

    NAME:           IOC9010_read( )

    Description:

    Function - 


    Parameters:

    Inputs:

    Outputs:
    Return:


    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

***********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 23-JAN-2006     DAN     1.1         Intial Version.
 26-MAR-2009	 JC		 1.2		 Totally rewritten. Modified to block/unblock read. Will be awaken by interrupt
 28-SEP-2009	 JC		 1.3		 Implemented to block/unblock read. Awaken by interrupt and copy the vector back to the user
 26-NOV-2010	 JC		 1.4		 When returning interrupt vector, it also returns the carrier id for multiple carrier systems.
 14-MAR-2011	 JC		 1.5		 Support for multiple files and multiple queues.
***********************************************************/
#undef MAX_NAMELEN 
#define MAX_NAMELEN 10
 
ssize_t IOC9010_read(struct file *filep, char __user *buff,  size_t count, loff_t *f_poss)
{
  	struct IOC9010_dev *ioc9010dev = filep->private_data;

/*	char localbuf[10];
 	int pending=0;
*/
	int i;
	
 	DEFINE_WAIT(wait);
	
 	if (count < sizeof(int))
 		return -EINVAL;
	
	/* loop all files and create queues for them */
	for(i=0; i<MAXFILENUM; i++)
	{
		if(ioc9010dev->irqdp[i].filep == filep)
		{
			ioc9010dev->irqdp[i].pending = atomic_read(&ioc9010dev->irqdp[i].atomCounter);
		 	while (ioc9010dev->irqdp[i].pending == 0) 
			{
	 			prepare_to_wait(&ioc9010dev->irqdp[i].queue, &wait, TASK_INTERRUPTIBLE);
				ioc9010dev->irqdp[i].pending = atomic_read(&ioc9010dev->irqdp[i].atomCounter);
				if (ioc9010dev->irqdp[i].pending == 0)
	 				schedule();
	 			finish_wait(&ioc9010dev->irqdp[i].queue, &wait);
	 			if (signal_pending(current))
	 				return -ERESTARTSYS;
 			}

			ioc9010dev->irqdp[i].buf[1] = ioc9010dev->devnum;
			ioc9010dev->irqdp[i].buf[0] = ioc9010dev->vector;
			ioc9010dev->vector = 0;								/* clear vector after it is sent */

			/* notify the user */
		 	if (copy_to_user(buff, ioc9010dev->irqdp[i].buf, 2))
		    {
				atomic_sub(ioc9010dev->irqdp[i].pending, &ioc9010dev->irqdp[i].atomCounter);
				return -EFAULT;
    		}
			atomic_sub(ioc9010dev->irqdp[i].pending, &ioc9010dev->irqdp[i].atomCounter);

/*	printk(KERN_NOTICE "%x Test - read called. filep=%x devNo=%d pend=%d\n", 
			ioc9010dev->DEVICE_ID, (int) filep, ioc9010dev->devnum, ioc9010dev->irqdp[i].pending);
*/
			break;
		}
	}
 
 	return 0;
}
 

/**********************************************************

    NAME:           IOC9010_write( )

    Description:

    Function - 

    Parameters:

    Inputs:

    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 23-JAN-2006     DAN     1.1         Intial Version.

***********************************************************/
ssize_t IOC9010_write(struct file *filep, const char __user *buff,  size_t count, loff_t *f_poss)
{
  	struct IOC9010_dev *ioc9010dev = filep->private_data;
	int status=0;

	/* Transfer to Kernel Space */
	status = copy_from_user((void *)(unsigned long)buffer, (void *)(unsigned long)buff, 2);

	/* Read From the IP Space */
	/* Quick Test - Read IP A's ID PROM */ 
	iowrite8(*(buffer++), ioc9010dev->VirtureMemBaseAddr + *f_poss); 
	iowrite8(*buffer, ioc9010dev->VirtureMemBaseAddr + *f_poss + 1); 
	buffer--;

	return status;
}


/**********************************************************

    NAME:           IOC9010_ioctl( )

    Description:

    Function - 


    Parameters:

    Inputs:

    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 23-JAN-2006     DAN     1.1         Intial Version.

***********************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
static int IOC9010_ioctl(struct inode *inode, struct file *file, unsigned int opcode, unsigned long param)
#else
static long IOC9010_ioctl(struct file *file, unsigned int opcode, unsigned long param)
#endif
{
	int 			iReturnValue;
	IOCTL_BUF		ioctl_buf;
	IOCTL_VME_BUF	ioctl_vme_buf;
	char			*cTransferBuffer;
	u16				counter=0;
	u16				IORead=0, *u16Transfer;
	u32				*u32Transfer;
	unsigned int	lTransferCount;
	unsigned int	addressOffset = 0;
	unsigned long	Address;
	unsigned long	lAddress;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,43))		//From Fedora 15. May need earlier version here?
	unsigned short  sdata1;
#endif
	unsigned short  sdata;
	unsigned short  sdataINT;								//for interrupt level setting
	int i, j, done=0;

  	struct IOC9010_dev *ioc9010dev = file->private_data;
  	
  	/* printk(KERN_NOTICE "HYTEC *** - IOCTL opcode = %x\n", (int)(opcode)); */ 
	
	/* this two line might be used for future implmentation */
	/* Set IP Site D as a 32 MHz Card */
	/* outw(CLK_IP_D_32MHZ, ioc9010dev->CarrierAddr + REG_IP_CLOCK); */

  	switch (opcode)
    	{
		case OP_ENABLE_INTERRUPT:
      		if (copy_from_user((void *)(unsigned long) &ioctl_vme_buf, (void *)(unsigned long)param, sizeof(ioctl_vme_buf)) > 0)
				return -EFAULT;
      		switch(ioc9010dev->DEVICE_ID)
      		{
      		case 0x5331:
      		case 0x5332:
        		/* Enable Interrupts for 5331/5332 */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))										//Before Fedora 15. May need earlier version here?
				sdata = inw(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);
				sdata |= 0x0043;
				outw(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);			//enable IRQ on 5331. added by Jim on 26-03-2009
				
				/* Enable interrupt for 3331 */
				sdata = inw(ioc9010dev->CarrierAddr + CSR);
				sdataINT = (unsigned short) ioctl_vme_buf.lLength;
				sdataINT |= 0x0010;
				sdata |= sdataINT;
    	        outw(sdata, ioc9010dev->CarrierAddr + CSR);								//set up IRQ7 and enable it on 3331
#else
				sdata = inb(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);
				sdata |= 0x43;
				outb(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);			//enable IRQ on 5331. added by Jim on 26-03-2009
				
				/* Enable interrupt for 3331 */
				sdata = inb(ioc9010dev->CarrierAddr + CSR);
				sdataINT = (unsigned short) ioctl_vme_buf.lLength;
				sdataINT |= 0x0010;
				sdata |= sdataINT;
    	        outb(sdata, ioc9010dev->CarrierAddr + CSR);								//set up IRQ7 and enable it on 3331
    	        
#endif
    	        break;
      		case 0x9010:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))										//Before Fedora 15. May need earlier version here?
				sdata = inw(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);
				sdata |= 0x043;
				outw(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);			//enable IRQ on PLX. added by Jim on 26-03-2009
				
        		/* Enable Carrier board Interrupts mask */
				sdata = inw(ioc9010dev->CarrierAddr + REG_IP_INT_MASK);
				sdata |= 1 << ioctl_vme_buf.lSite * 2;									//shift to the right bit. Note: Hytec IP only uses INT0
				outw(sdata, ioc9010dev->CarrierAddr + REG_IP_INT_MASK); 				//enable the IP interrupt
    			outw(0x0000, ioc9010dev->CarrierAddr + REG_IP_ERROR_MASK);				//01/12/2010 JC, this causes grief when IP card not pull it high. So set it to 0!
#else
				sdata = inb(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);
				sdata |=  0x43;
				outb(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);			//enable IRQ on PLX. added by Jim on 26-03-2009

				/* Enable Carrier board Interrupts mask */
				sdata = inb(ioc9010dev->CarrierAddr + REG_IP_INT_MASK);
				sdata1 = inb(ioc9010dev->CarrierAddr + REG_IP_INT_MASK + 1);
				if(ioctl_vme_buf.lSite < 4)
				{
					sdata |= 1 << (ioctl_vme_buf.lSite * 2);							//shift to the right bit. Note: Hytec IP only uses INT0
					outb(sdata, ioc9010dev->CarrierAddr + REG_IP_INT_MASK); 			//enable the IP interrupt
				}else
				{
					sdata1 |= 1 << ((ioctl_vme_buf.lSite - 4) * 2);						//shift to the right bit. Note: Hytec IP only uses INT0
					outb(sdata1, ioc9010dev->CarrierAddr + REG_IP_INT_MASK + 1); 		//enable the IP interrupt
				}
    			outb(0x00, ioc9010dev->CarrierAddr + REG_IP_ERROR_MASK);				//01/12/2010 JC, this causes grief when IP card not pull it high. So set it to 0!
    			outb(0x00, ioc9010dev->CarrierAddr + REG_IP_ERROR_MASK + 1);			//01/12/2010 JC, this causes grief when IP card not pull it high. So set it to 0!
#endif    	      			
      			break;
      		case 0x7003:
				sdata = ioread16(ioc9010dev->VirtureCarrierRegBaseAddr + Hytec7002CSR);
				sdata |= 0x0001;
				iowrite16(sdata, ioc9010dev->VirtureCarrierRegBaseAddr + Hytec7002CSR);   
				
        		/* Enable Carrier board Interrupts mask */
				sdata = ioread16(ioc9010dev->VirtureCarrierRegBaseAddr + REG_IP_INT_MASK);
				sdata |= 1 << ioctl_vme_buf.lSite * 2;									//shift to the right bit. Note: Hytec IP only uses INT0
				iowrite16(sdata, ioc9010dev->VirtureCarrierRegBaseAddr + REG_IP_INT_MASK);
				iowrite16(0x0000, ioc9010dev->VirtureCarrierRegBaseAddr + REG_IP_ERROR_MASK);
      			break;
      		case 0x7002:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))										//Before Fedora 15. May need earlier version here?
				sdata = inw(ioc9010dev->CarrierAddr + Hytec7002CSR);
				sdata |= 0x0001;
				outw(sdata, ioc9010dev->CarrierAddr + Hytec7002CSR);					//enable IRQ in FPGA. added by Jim on 13-10-2010
#else
				sdata = inb(ioc9010dev->CarrierAddr + Hytec7002CSR);
				sdata |= 0x01;
				outb(sdata, ioc9010dev->CarrierAddr + Hytec7002CSR);					//enable IRQ in FPGA. added by Jim on 13-10-2010
#endif    	      
      			break;
      		case 0x6335:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))										//Before Fedora 15. May need earlier version here?
				sdata = inw(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR8311);
				sdata |= 0x0900;
				outw(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR8311);		//enable IRQ on PLX. added by Jim on 26-03-2009

				/* Enable Carrier board Interrupts mask */
				sdata = inw(ioc9010dev->CarrierAddr + REG_IP_INT_MASK);
				sdata |= 1 << ioctl_vme_buf.lSite * 2;									//shift to the right bit. Note: Hytec IP only uses INT0
    			outw(sdata, ioc9010dev->CarrierAddr + REG_IP_INT_MASK); 				//enable the IP interrupt
        		outw(0x0000, ioc9010dev->CarrierAddr + REG_IP_ERROR_MASK);				//01/12/2010 JC, this causes grief when IP card not pull it high. So set it to 0!
#else
				sdata = inb(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR8311 + 1);
				sdata |=  0x09;
				outb(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR8311 + 1);	//enable IRQ on PLX. added by Jim on 26-03-2009
				sdata = inb(ioc9010dev->CarrierAddr + REG_CSR);
				sdata |= 0x0001;
				outb(sdata, ioc9010dev->CarrierAddr + REG_CSR);							//enable IRQ in FPGA. added by Jim on 13-10-2010

				/* Enable Carrier board Interrupts mask */
				sdata = inb(ioc9010dev->CarrierAddr + REG_IP_INT_MASK);
				sdata |= 1 << (ioctl_vme_buf.lSite * 2);								//shift to the right bit. Note: Hytec IP only uses INT0
				outb(sdata, ioc9010dev->CarrierAddr + REG_IP_INT_MASK); 			//enable the IP interrupt
        		outb(0x00, ioc9010dev->CarrierAddr + REG_IP_ERROR_MASK);			//01/12/2010 JC, this causes grief when IP card not pull it high. So set it to 0!
#endif    	      
      			break;
      		}
			break;

		case OP_DISABLE_INTERRUPT:														/* Disable Interrupts */
			switch (ioc9010dev->DEVICE_ID)
            {
            case 0x9010:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))										//Before Fedora 15. May need earlier version here?
				sdata = inw(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);
				sdata &= (~0x040);
				outw(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);			//clear IRQ on PLX. added by Jim on 26-03-2009
#else
				sdata = inb(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);
				sdata &= (~0x40);
				outb(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);			//clear IRQ on PLX. added by Jim on 26-03-2009
#endif    	      
				break;
            case 0x6335:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))										//Before Fedora 15. May need earlier version here?
				sdata = inw(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR8311);
				sdata &= (~0x0900);
				outw(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR8311);		//clear IRQ on PLX. added by Jim on 26-03-2009
#else
				sdata = inb(ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR8311 + 1);
				sdata &= (~0x09);
				outb(sdata, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR8311 + 1);	//clear IRQ on PLX. added by Jim on 26-03-2009						
#endif    	      
				break;
            case 0x7002:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))										//Before Fedora 15. May need earlier version here?
				sdata = inw(ioc9010dev->CarrierAddr + Hytec7002CSR);
				sdata &= (~0x0001);
				outw(sdata, ioc9010dev->CarrierAddr + Hytec7002CSR);					//clear IRQ in FPGA. added by Jim on 13-10-2010
#else
				sdata = inb(ioc9010dev->CarrierAddr + Hytec7002CSR);
				sdata &= (~0x01);
				outb(sdata, ioc9010dev->CarrierAddr + Hytec7002CSR);					//clear IRQ in FPGA. added by Jim on 13-10-2010
#endif    	      
				break;
            case 0x7003:
				sdata = ioread16(ioc9010dev->VirtureCarrierRegBaseAddr + Hytec7002CSR);
				sdata &= (~0x0001);
				iowrite16(sdata, ioc9010dev->VirtureCarrierRegBaseAddr + Hytec7002CSR);
				break;
				
			case 0x5331:
            case 0x5332:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))			//Before Fedora 15. May need earlier version here?
			    outw(0x0003, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);
#else
			    outb(0x03, ioc9010dev->PLXLocalConfigBAR + PLXLocalConfigCSR);    	        
#endif    	      
				break;
			}

		case OP_BASE_ADD3:
			/* Get Parameters Structure */  
/*       		if (copy_from_user((void *)(unsigned long)&ioctl_buf, (void *)(unsigned long)param, sizeof(ioctl_buf)) > 0)
		    return -EFAULT;
*/
    	  	/* For Moment Only Information in Structure is Hardware ID */    
            switch (ioc9010dev->DEVICE_ID)
            {
                case 0x9010:
                case 0x6335:
                case 0x7002:
                case 0x7003:
                    ioctl_buf.lAddress = ioc9010dev->IOCMemoryAddr;
                    break;
            	case 0x5331:
                case 0x5332:
                    ioctl_buf.lAddress = ioc9010dev->CarrierAddr;
                    break;
            }       
            
      		if (copy_to_user((void *)(unsigned long)param, (void *)(unsigned long)&ioctl_buf, sizeof(ioctl_buf)) > 0)
      		{
      			printk(KERN_NOTICE "HYTEC 9010/6335 - Try to copy base3 to user error=%d. \n", done);    	
        		return -EFAULT;
    		}    
			break;

		case OP_BASE_ADD2:
			/* Get Parameters Structure */  
       		if (copy_from_user((void *)(unsigned long)&ioctl_buf, (void *)(unsigned long)param, sizeof(ioctl_buf)) > 0)
        	return -EFAULT;

    	  	/* For Moment Only Information in Structure is Hardware ID */    
            switch (ioc9010dev->DEVICE_ID)
            {
                case 0x9010:
                case 0x6335:
                case 0x7002:
                case 0x7003:
                    ioctl_buf.lAddress = ioc9010dev->CarrierAddr;
                    break;
            	case 0x5331:
                case 0x5332:
                    ioctl_buf.lAddress = 0;
                    break;
            }       
            
      		if (copy_to_user((void *)(unsigned long)param, (void *)(unsigned long)&ioctl_buf, sizeof(ioctl_buf)) > 0)
      		{
      			printk(KERN_NOTICE "HYTEC 9010/6335 - Try to copy base2 to user. \n");    	
        		return -EFAULT;
    		}
    
			break;

		case OP_BASE_ADD1:
			/* Get Parameters Structure */  
       		if (copy_from_user((void *)(unsigned long)&ioctl_buf, (void *)(unsigned long)param, sizeof(ioctl_buf)) > 0)
        	return -EFAULT;

    	  	/* For Moment Only Information in Structure is Hardware ID */    
            switch (ioc9010dev->DEVICE_ID)
            {
                case 0x9010:
                case 0x6335:
					ioctl_buf.lAddress = ioc9010dev->PLXLocalConfigBAR;
                    break;
                case 0x7002:
                case 0x7003:
                    ioctl_buf.lAddress = ioc9010dev->CarrierAddr;
                    break;
            	case 0x5331:
                case 0x5332:
                    ioctl_buf.lAddress = ioc9010dev->PLXLocalConfigBAR;
                    break;
            }       
            
      		if (copy_to_user((void *)(unsigned long)param, (void *)(unsigned long)&ioctl_buf, sizeof(ioctl_buf)) > 0)
      		{
      			printk(KERN_NOTICE "HYTEC 9010/6335 - Try to copy base1 to user. \n");    	
        		return -EFAULT;
    		}
    
			break;

		case OP_REGISTER_PROCESSID:
			/* Get Parameters Structure */  
      		if (copy_from_user((void *)(unsigned long)&ioctl_buf, (void *)(unsigned long)param, sizeof(ioctl_buf)) > 0)
        	return -EFAULT;

            ioc9010dev->pid[ioc9010dev->pidNumber] = ioctl_buf.sData;             // store user PIDs
            ioc9010dev->pidNumber++;
            if (ioc9010dev->pidNumber >= TASK_MAXPID) return -EFAULT;        // cannot exceed maximum pid number
            ioc9010dev->signal = ioctl_buf.lAddress;                         // store user specified signal
    
			break;

		case OP_REGISTER_VECTOR:
      		if (copy_from_user((void *)(unsigned long)&ioctl_buf, (void *)(unsigned long)param, sizeof(ioctl_buf)) > 0)
        	return -EFAULT;

			done = 0;
			/* register vector for the file */
			for(i=0; i<MAXFILENUM; i++)
			{
				if(ioc9010dev->irqdp[i].filep == file)
				{
					for(j=0; j<MAXIPNUM; j++)
					{
						if(ioc9010dev->irqdp[i].vector[j] == 0)
						{
							ioc9010dev->irqdp[i].vector[j] = (char) ioctl_buf.sData & 0xFF;
							done = 1;
							break;			//for MAXIPNUM
						}
					}
					if(done == 1) break;	//for MAXFILENUM
				}
			}
			break;

		case OP_SET_5331_CSR:	/* Set CSR on 5331/5332 Only. added by Jim on 26-03-2009 */       	
			/* Get Parameters Structure */  
      		if (copy_from_user((void *)(unsigned long)&sdata, (void *)(unsigned long)param, sizeof(sdata)) > 0)
        		return -EFAULT;

			outw(sdata, ioc9010dev->CarrierAddr + CSR);		//write WORD to 3331 CSR
			break;

		case OP_GET_CONFIG:

			/* Get Parameters Structure */  
       		if (copy_from_user((void *)(unsigned long)&ioctl_buf, (void *)(unsigned long)param, sizeof(ioctl_buf)) > 0)
        	return -EFAULT;

    	  	/* For Moment Only Information in Structure is Hardware ID */    
            	switch (ioc9010dev->DEVICE_ID)
            	{
                	case 0x9010:
                	case 0x6335:
                		ioctl_buf.sData = ioc9010dev->DEVICE_ID;
                   	 break;
                	case 0x7002:
                	case 0x7003:
                    	ioctl_buf.sData = ioc9010dev->DEVICE_ID;
                    	break;
            		case 0x5331:
                	case 0x5332:
                    	ioctl_buf.sData = ioc9010dev->DEVICE_ID;
                    	break;
            	}       
            
      		if (copy_to_user((void *)(unsigned long)param, (void *)(unsigned long)&ioctl_buf, sizeof(ioctl_buf)) > 0)
      		{
      			printk(KERN_NOTICE "HYTEC 9010/6335 - Try to copy base1 to user. \n");    	
        		return -EFAULT;
    		}

			break;

		case OP_GET_5331_CSR:
            
            IORead = inw(ioc9010dev->CarrierAddr+CSR);

      		if (copy_to_user((void *)(unsigned long)param, (void *)(unsigned long)&IORead, sizeof(IORead)) > 0)
        	return -EFAULT;

			break;

		case OP_3331_RESET:
            
			/* Initiate the RESET */
			outw(0x0080, ioc9010dev->CarrierAddr+CSR);

			/* wait for the reset to activate */
			IORead = 0x80;
			counter = 0;
			while ((IORead & 0x0080) && (++counter<10))
			{
	            IORead = inw(ioc9010dev->CarrierAddr+CSR);
				msleep(50);
			}

			if(IORead & 0x0080) 
        	return -EFAULT;

			/* wait for the reset to complete */
			outw(0x0001, ioc9010dev->CarrierAddr+CSR);
            IORead = inw(ioc9010dev->CarrierAddr+CSR);
			while ((!(IORead & 0x0001)) && (++counter<15))
			{
				msleep(50);
				outw(0x0001, ioc9010dev->CarrierAddr+CSR);
	            IORead = inw(ioc9010dev->CarrierAddr+CSR);
			}

			if(!IORead) 
        	return -EFAULT;

			break;

		case OP_GENERAL_VME_READ_BLOCK:
			/* Get Parameters Structure */  
      		if (copy_from_user((void *)(unsigned long)&ioctl_vme_buf, (void *)(unsigned long)param, sizeof(ioctl_vme_buf)) > 0)
        	return -EFAULT;

			/* Allocate the Kernal space Buffer  */
			u16Transfer = kmalloc(ioctl_vme_buf.lLength, GFP_KERNEL);

			dev_lock;			/* spinlock code */
            /* If Access is to IO/ID Space (i.e. NOT Memory) */            
            if (ioctl_vme_buf.lMemAccess == FALSE)
            {
			    /* Work out the full address for 5331 PCI Card */
                Address = (ioctl_vme_buf.lSlot << 11)
                        + (ioctl_vme_buf.lSite << 8)
                        +  ioctl_vme_buf.lAddress;    

			    /* Read the Data via the 5331 PCI Card */
			    vme_block_read_short_16(ioc9010dev, 
										Address,
				    					u16Transfer, 
				    					ioctl_vme_buf.lLength/2);
            }
            /* If Access is to Memory Space */            
            else
            {
			    /* Work out the full address for 5331 PCI Card */
                lAddress = (long)(ioctl_vme_buf.lSlot << 22)
                         + (long)(ioctl_vme_buf.lSite << 20)
                         + (long)(ioctl_vme_buf.lAddress + (0x01<<28));   


			    /* Read the Data via the 5331 PCI Card */
			    vme_block_read_long_16(ioc9010dev,
									   lAddress,
				    				   u16Transfer, 
				    				   ioctl_vme_buf.lLength/2);
            }
			dev_unlock;			/* spinlock code */

			/* Copy Data Read in the Kernal Buffer to User Space */
            /* - sData is a pointer to the Data in User Space */  
      		if (copy_to_user((void *)(unsigned long)ioctl_vme_buf.sData, (void *)(unsigned long)u16Transfer, ioctl_vme_buf.lLength)>0)
        	return -EFAULT;

			/* Free Memory Allocated to Kernel buffer */
			kfree(u16Transfer);

			break;

		case OP_GENERAL_VME_READ_BLOCK_32:
			/* Get Parameters Structure */  
      		if (copy_from_user((void *)(unsigned long)&ioctl_vme_buf, (void *)(unsigned long)param, sizeof(ioctl_vme_buf)) > 0)
        	return -EFAULT;

			/* Allocate the Kernal space Buffer  */
			u32Transfer = kmalloc(ioctl_vme_buf.lLength, GFP_KERNEL);

			dev_lock;			/* spinlock code */
            /* If Access is to IO/ID Space (i.e. NOT Memory) */            
            if (ioctl_vme_buf.lMemAccess == FALSE)
            {
			    /* Work out the full address for 5331 PCI Card */
                Address = (ioctl_vme_buf.lSlot << 11)
                        + (ioctl_vme_buf.lSite << 8)
                        +  ioctl_vme_buf.lAddress;    

			    /* Read the Data via the 5331 PCI Card */
			    vme_block_read_long_32( ioc9010dev,
										Address,
				    					(long *)(u32Transfer),
				    					ioctl_vme_buf.lLength/4);
            }
            /* If Access is to Memory Space */            
            else
            {
			    /* Work out the full address for 5331 PCI Card */
                lAddress = (long)(ioctl_vme_buf.lSlot << 22)
                         + (long)(ioctl_vme_buf.lSite << 20)
                         + (long)(ioctl_vme_buf.lAddress + (0x01<<28));   


			    /* Read the Data via the 5331 PCI Card */
			    vme_block_read_long_32(ioc9010dev,
									   lAddress,
				    				   (long *)(u32Transfer), 
				    				   ioctl_vme_buf.lLength/4);
            }
			dev_unlock;			/* spinlock code */

			/* Copy Data Read into the Kernal Buffer to User Space */
            /* - sData is a pointer to the Data in User Space */  
      		if (copy_to_user((void *)(unsigned long)ioctl_vme_buf.sData, (void *)(unsigned long)u32Transfer, ioctl_vme_buf.lLength)>0)
        	return -EFAULT;

			/* Free Memory Allocated to Kernel buffer */
			kfree(u32Transfer);

			break;


		case OP_GENERAL_VME_WRITE_BLOCK:
			/* Get Parameters Structure */  
      		if (copy_from_user((void *)(unsigned long)&ioctl_vme_buf, (void *)(unsigned long)param, sizeof(ioctl_vme_buf)) > 0)
        	return -EFAULT;

			/* Allocate the Kernal space Buffer  */
			u16Transfer = kmalloc(ioctl_vme_buf.lLength, GFP_KERNEL);

			/* Get Data to Write into the Kernal Buffer */
            /* - sData is a pointer to the Data in User Space */  
      		if (copy_from_user( (void *)(unsigned long)u16Transfer,
                                (void *)(unsigned long)ioctl_vme_buf.sData,
                                ioctl_vme_buf.lLength) > 0)
        	return -EFAULT;

			dev_lock;			/* spinlock code */
            /* If Access is to IO/ID Space (i.e. NOT Memory) */            
            if (ioctl_vme_buf.lMemAccess == FALSE)
            {
			    /* Work out the full address for 5331 PCI Card */
	            Address = (ioctl_vme_buf.lSlot << 11)
		                 + (ioctl_vme_buf.lSite << 8)
			             +  ioctl_vme_buf.lAddress;    

				/* Write the Data via the 5331 PCI Card */
				vme_block_write_short_16(ioc9010dev,
										 Address,
								 		 u16Transfer, 
										 ioctl_vme_buf.lLength/2);
            }
            /* If Access is to Memory Space */            
            else
            {
			    /* Work out the full address for 5331 PCI Card */
                lAddress = (long)(ioctl_vme_buf.lSlot << 22)
                         + (long)(ioctl_vme_buf.lSite << 20)
                         + (long)(ioctl_vme_buf.lAddress + (0x01<<28));   

			    /* Write the Data via the 5331 PCI Card */
			    vme_block_write_long_16(ioc9010dev,
										lAddress,
				    				    u16Transfer, 
				    				    ioctl_vme_buf.lLength/2);
            }
			dev_unlock;			/* spinlock code */

			/* Free Memory Allocated to Kernel buffer */
			kfree(u16Transfer);

			break;

		case OP_GENERAL_VME_WRITE_BLOCK_32:

			/* Get Parameters Structure */  
      		if (copy_from_user((void *)(unsigned long)&ioctl_vme_buf, (void *)(unsigned long)param, sizeof(ioctl_vme_buf)) > 0)
        	return -EFAULT;

			/* Allocate the Kernal space Buffer  */
			u32Transfer = kmalloc(ioctl_vme_buf.lLength, GFP_KERNEL);

			/* Get Data to Write into the Kernal Buffer */
            /* - sData is a pointer to the Data in User Space */  
      		if (copy_from_user( (void *)(unsigned long)u32Transfer,
                                (void *)(unsigned long)ioctl_vme_buf.sData,
                                ioctl_vme_buf.lLength) > 0)
        	return -EFAULT;

			dev_lock;			/* spinlock code */
            /* If Access is to IO/ID Space (i.e. NOT Memory) */            
            if (ioctl_vme_buf.lMemAccess == FALSE)
            {
			    /* Work out the full address for 5331 PCI Card */
	            Address = (ioctl_vme_buf.lSlot << 11)
		                 + (ioctl_vme_buf.lSite << 8)
			             +  ioctl_vme_buf.lAddress;    

                /* Write the Data via the 5331 PCI Card */
				vme_block_write_long_32(ioc9010dev,
										Address,
								 		(long *)u32Transfer, 
										ioctl_vme_buf.lLength/4);
            }
            /* If Access is to Memory Space */            
            else
            {
			    /* Work out the full address for 5331 PCI Card */
                lAddress = (long)(ioctl_vme_buf.lSlot << 22)
                         + (long)(ioctl_vme_buf.lSite << 20)
                         + (long)(ioctl_vme_buf.lAddress + (0x01<<28));   

			    /* Write the Data via the 5331 PCI Card */
			    vme_block_write_long_32(ioc9010dev,
										lAddress,
				    				    (long *)u32Transfer, 
				    				    ioctl_vme_buf.lLength/4);
            }
			dev_unlock;			/* spinlock code */

			/* Free Memory Allocated to Kernel buffer */
			kfree(u32Transfer);

			break;
    
		/* General Read */
		case OP_GENERAL_READ:

			/* Get Parameters Structure */  
      		if (copy_from_user((void *)(unsigned long)&ioctl_buf, (void *)(unsigned long)param, sizeof(ioctl_buf)) > 0)
        	return -EFAULT;

    	  	/* Really Read Hardware */    
            iReturnValue = ioread16(ioc9010dev->VirtureMemBaseAddr + ioctl_buf.lAddress);
            
      		if (copy_to_user((void *)(unsigned long)param, (void *)(unsigned long)&ioctl_buf, sizeof(ioctl_buf)) > 0)
        	return -EFAULT;

    		return iReturnValue;

		/* General Read Block of Data from IP IO and ID Region */
		/* IMPORTANT NOTE - No Break so it falls through to the next switch */
		case OP_GENERAL_READ_BLOCK:

			/* Modify the Address Offset from 0 (default) to point */
            /* it at IO Region of the IP Card Space */
			addressOffset = IP_IO_BASE_ADDR;

		/* General Read Block of Data from IP Memory Region */
		case OP_GENERAL_MEM_READ_BLOCK:

			/* Get Parameters Structure */  
      		if (copy_from_user((void *)(unsigned long)&ioctl_buf, (void *)(unsigned long)param, sizeof(ioctl_buf)) > 0)
        	return -EFAULT;

			/* Allocate the Kernal space Buffer  */
			cTransferBuffer = kmalloc(ioctl_buf.lLength * 2, GFP_KERNEL);

			dev_lock;			/* spinlock code */
			/* Read the Data to the PCI/IP Card Space */
    	  	for (lTransferCount=0; lTransferCount<(ioctl_buf.lLength); lTransferCount+=2)
			{
                IORead = ioread16(ioc9010dev->VirtureMemBaseAddr + addressOffset + ioctl_buf.lAddress);

				*(cTransferBuffer + lTransferCount    ) = (u8)(IORead & 0x00FF);
				*(cTransferBuffer + lTransferCount + 1) = (u8)(IORead >> 8);
				ioctl_buf.lAddress+=2;
			}
			dev_unlock;			/* spinlock code */

			/* Copy Data Read into the Kernal Buffer to User Space */
            /* - sData is a pointer to the Data in User Space */  
      		if (copy_to_user((void *)(unsigned long)ioctl_buf.sData, (void *)(unsigned long)cTransferBuffer, ioctl_buf.lLength*2) > 0)
        	return -EFAULT;


			/* Free Memory Allocated to Kernel buffer */
			kfree(cTransferBuffer);

			break;


		/* General Write Block of Data from IP IO and ID Region */
		/* IMPORTANT NOTE - No Break so it falls through to the next switch */
		case OP_GENERAL_WRITE_BLOCK:

			/* Modify the Address Offset from 0 (default) to point */
            /* it at IO Region of the IP Card Space */
			addressOffset = IP_IO_BASE_ADDR;

		/* General Write Block of Data from IP Memory Region */
		case OP_GENERAL_MEM_WRITE_BLOCK:

			/* Get Parameters Structure */  
      		if (copy_from_user((void *)(unsigned long)&ioctl_buf, (void *)(unsigned long)param, sizeof(ioctl_buf)) > 0)
        	return -EFAULT;

			/* Allocate the Kernal space Buffer  */
			cTransferBuffer = kmalloc(ioctl_buf.lLength * 2, GFP_KERNEL);

			/* Get Data to Write into the Kernal Buffer */
            /* - sData is a pointer to the Data in User Space */  
      		if (copy_from_user((void *)(unsigned long)cTransferBuffer,(void *)(unsigned long)ioctl_buf.sData,ioctl_buf.lLength*2) > 0)
        	return -EFAULT;

			dev_lock;			/* spinlock code */
			/* Write the Data to the PCI/IP Card Space */
  	  		for (lTransferCount=0; lTransferCount<(ioctl_buf.lLength); lTransferCount+=2)
			{
				u16Transfer = (u16*)(cTransferBuffer + lTransferCount);

				iowrite16(*u16Transfer, ioc9010dev->VirtureMemBaseAddr + addressOffset + ioctl_buf.lAddress);

				ioctl_buf.lAddress+=2;
			}
			dev_unlock;			/* spinlock code */
		
			/* Free Memory Allocated to Kernel buffer */
			kfree(cTransferBuffer);

			break;

		/* Read Block of Data off IOC9010 Carrier Board */
		case OP_CARRIER_READ_BLOCK:

			/* Get Parameters Structure */  
      		if (copy_from_user((void *)(unsigned long)&ioctl_buf, (void *)(unsigned long)param, sizeof(ioctl_buf)) > 0)
        	return -EFAULT;

			/* Allocate the Kernal space Buffer  */
			cTransferBuffer = kmalloc(ioctl_buf.lLength * 2, GFP_KERNEL);

			dev_lock;			/* spinlock code */
			/* Read Data from the PCI/IP Space into the Kernal Buffer */
    	  	for (lTransferCount = 0; lTransferCount < (ioctl_buf.lLength * 2); lTransferCount+=2)
			{
				if(ioc9010dev->DEVICE_ID == 0x7003)
					IORead = ioread16(ioc9010dev->VirtureCarrierRegBaseAddr + ioctl_buf.lAddress + lTransferCount);
				else
					IORead = inw(ioc9010dev->CarrierAddr + ioctl_buf.lAddress++);

                *(cTransferBuffer + lTransferCount)     = IORead & 0xFF;
        		*(cTransferBuffer + lTransferCount + 1) = IORead >> 8;
			}
			dev_unlock;			/* spinlock code */

			/* Copy Data Read into the Kernal Buffer to User Space */
            /* - sData is a pointer to the Data in User Space */  
      		if (copy_to_user((void *)(unsigned long)ioctl_buf.sData, (void *)(unsigned long)cTransferBuffer, ioctl_buf.lLength*2) > 0)
        	return -EFAULT;
			/* Free Memory Allocated to Kernel buffer */
			kfree(cTransferBuffer);

			break;

		/* Write Block of Data to the IOC9010 Carrier Board */
		case OP_CARRIER_WRITE_BLOCK:

			/* Get Parameters Structure */  
      		if (copy_from_user((void *)(unsigned long)&ioctl_buf, (void *)(unsigned long)param, sizeof(ioctl_buf)) > 0)
        	return -EFAULT;

			/* Allocate the Kernal space Buffer  */
			cTransferBuffer = kmalloc(ioctl_buf.lLength * 2, GFP_KERNEL);

			/* Get Data to Write into the Kernal Buffer */
            /* - sData is a pointer to the Data in User Space */  
      		if (copy_from_user((void *)(unsigned long)cTransferBuffer,(void *)(unsigned long)ioctl_buf.sData,ioctl_buf.lLength*2) > 0)
        	return -EFAULT;

			dev_lock;			/* spinlock code */
			/* Write Data to the PCI/IP Space into the Kernal Buffer */
    	  	for (lTransferCount = 0; lTransferCount < (ioctl_buf.lLength * 2); lTransferCount+=2)
			{
                IORead =  (*(cTransferBuffer + lTransferCount) & 0xFF);
        		IORead |= (*(cTransferBuffer + lTransferCount + 1) << 8);

				if(ioc9010dev->DEVICE_ID == 0x7003)
					iowrite16(IORead, ioc9010dev->VirtureCarrierRegBaseAddr + ioctl_buf.lAddress + lTransferCount);
				else
					outw(IORead, ioc9010dev->CarrierAddr + ioctl_buf.lAddress++);
			}
			dev_unlock;			/* spinlock code */

			/* Free Memory Allocated to Kernel buffer */
			kfree(cTransferBuffer);

			break;
	}

  	/* signal success */
  	return 1;
}
 

struct file_operations IOC9010_fops = {
        .owner   = THIS_MODULE,
        .open    = IOC9010_open,
        .read    = IOC9010_read,
        .write   = IOC9010_write,
        .mmap    = IOC9010_mmap,                // for memory mapping 
        #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
        .ioctl          = IOC9010_ioctl,
        #else
        .unlocked_ioctl = (void*)(long)IOC9010_ioctl,
        #endif
/*        .poll    = IOC9010_poll,*/
        .release = IOC9010_release,
        .fasync  = IOC9010_fasync,   // for interrupt asynchronous notification
};


/**********************************************************

    NAME:           probe( )

    Description:

    Function - 


    Parameters:

    Inputs:
    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 23-JAN-2006     DAN     1.1         Intial Version.
 26-MAR-2009	 JC		 1.2		 Added interrupt support for all:9010/5331/5332/6335
									 Added support for PCI Express 6335
 22-NOV-2010	 JC		 1.3		 Remove device number registration to init
									 Added multiple device support
***********************************************************/
static int probe(struct pci_dev *dev, const struct pci_device_id *id)
{
	u16 iProbeVariable;
	u32 iBaseAddressVariable;
	u32 iCarrierBoardBaseAddress;

	u32 IPOffset;
	u16 DataRead, sbuf[20];
	u16 iCardCount = 0;
	//u16 data_w = 0;

	char String0[40];
	char String1[40];

	int result, err, i, j;

	struct IOC9010_dev* ioc9010dev = devlist; 
	struct IOC9010_dev* devtemp; 

	int serialNum = 0, temp = 0;

	/* Begin */
	printk(KERN_NOTICE "HYTEC 9010/5331/5332/6335/7002/7003 - probe called \n");

	/* Create a device and added to the device chain */
	ioc9010dev = kmalloc(sizeof(struct IOC9010_dev), GFP_KERNEL);
	if(ioc9010dev == NULL)
	{
		printk(KERN_NOTICE "HYTEC 9010/5331/5332/6335/7002/7003 - memory allocationi error. \n");
		return 0;
	}
	memset(ioc9010dev, 0, sizeof(struct IOC9010_dev));

	ioc9010dev->next = devlist;
	devtemp = devlist;
	devlist = ioc9010dev;

	/****** Interrupt stuff added by Jim on 26-03-2009 */
	ioc9010dev->vector = 0;								/* clear interrupt vector. */
    ioc9010dev->pidNumber = 0;                       	/* clear pid registration */
    ioc9010dev->signal = SIGIO;                      	/* default signal for interrupt notification */

  	/* init spinlock areas */
	spin_lock_init(&ioc9010dev->device_lock);

	// init states
	ioc9010dev->irqState = 0;
	ioc9010dev->bar1State = 0;
	ioc9010dev->bar2State = 0;
	ioc9010dev->bar3State = 0;

	/* initialise file array */
	for(i=0; i<MAXFILENUM; i++)
	{
		ioc9010dev->irqdp[i].filep = NULL;
		ioc9010dev->irqdp[i].pending = 0;
		for(j=0; j<MAXIPNUM; j++)
			ioc9010dev->irqdp[i].vector[j] = 0;
	}

	/* Device Registration */
	cdev_init(&ioc9010dev->cdev, &IOC9010_fops);
	ioc9010dev->cdev.owner = THIS_MODULE;
	ioc9010dev->cdev.ops = &IOC9010_fops;

	/* Enable PCI Device */
	result = pci_enable_device(dev);
	if (result != 0)
	{
		printk(KERN_NOTICE "HYTEC 9010/5331/5332/6335/7002/7003 - Initial PCI device error. %x\n", result);
		return(-1);
	}
	
	/* Check if its a Hytec Product */
	pci_read_config_word(dev, PCI_VENDOR_ID, &iProbeVariable);
	printk(KERN_NOTICE "HYTEC 9010/5331/5332/6335/7002/7003 - Vendor ID - 0x%4X \n", iProbeVariable);

	ioc9010dev->VENDOR_ID = iProbeVariable;
	
	/* Check if its a 9010 IOC Blade */
	/* ----------------------------- */
	pci_read_config_word(dev, PCI_DEVICE_ID, &iProbeVariable);
	printk(KERN_NOTICE "HYTEC 9010/5331/5332/6335/7002/7003 - Device ID - 0x%4X \n", iProbeVariable);

	ioc9010dev->DEVICE_ID = iProbeVariable;

    switch (ioc9010dev->DEVICE_ID)
    {
    case 0x9010:
    case 0x6335:
    case 0x7002:
    case 0x7003:

    	/* Map the Base Address for the 9010's PCI Carrier Card Registers */
    	/* -------------------------------------------------------------- */
    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_2, &iCarrierBoardBaseAddress);
		
		/* Update Log about it all went */
    	printk(KERN_NOTICE "%x Test - Base Add2 - 0x%8X \n",ioc9010dev->DEVICE_ID, iCarrierBoardBaseAddress);
		if (ioc9010dev->DEVICE_ID == 0x7003)													// for 7003, carrier registers are mapped to memory 
		{
			if (!request_mem_region(iCarrierBoardBaseAddress, 64, "Hytec Carrier Registers"))
			{
				printk(KERN_NOTICE "%x Test - Carrier memory region mapping failed!\n", ioc9010dev->DEVICE_ID);
			}
			else
			{
				printk(KERN_NOTICE "%x Test - Carrier memory mapped !\n", ioc9010dev->DEVICE_ID);
				ioc9010dev->bar2State = 1;														//BAR3 mapped successfully
			}

			ioc9010dev->VirtureCarrierRegBaseAddr = ioremap(iCarrierBoardBaseAddress, 64);
		}else
		{
	    	/* These are I/O, hence LSB is only set to indicate an I/O Region */
    		iCarrierBoardBaseAddress--;			//bit0 = 1 means it is I/O base, so mask bit0

			if (!request_region(iCarrierBoardBaseAddress, 64, "Hytec Carrier Registers"))
			{
				printk(KERN_NOTICE "%x Test - Carrier region is also used by another driver !\n", ioc9010dev->DEVICE_ID);
			}
			else

			{
				printk(KERN_NOTICE "%x Test - Carrier Mapped !\n", ioc9010dev->DEVICE_ID);
				ioc9010dev->bar2State = 1;														//BAR2 mapped successfully
			}
		}
		ioc9010dev->CarrierAddr = iCarrierBoardBaseAddress;										//save it for releasing


		if (ioc9010dev->DEVICE_ID == 0x7003)													/* get the serial number of the 7003 carrier */
		{
			serialNum = ioread16(ioc9010dev->VirtureCarrierRegBaseAddr + 0x04);
			while (devtemp != NULL)
			{
				if (devtemp->devnum == serialNum)
				{
					printk(KERN_NOTICE "%x Test - Duplicated carrier ID %d found (Not registered). Please check carrier ID jumpers.\n", devtemp->DEVICE_ID, serialNum);
					devlist = ioc9010dev->next;
					kfree(ioc9010dev);							/* free the current */
					return(-1);
				}
				devtemp = devtemp->next;
			}
			dtMajorMinor = MKDEV(ioc_major, ioc_minor + serialNum);
			printk(KERN_NOTICE "%x Test - Carrier ID=%d.\n", ioc9010dev->DEVICE_ID, serialNum);

			temp = ioread16(ioc9010dev->VirtureCarrierRegBaseAddr + 0x14);
			printk(KERN_NOTICE "%x Test - Carrier card model = %x.\n", ioc9010dev->DEVICE_ID, temp);
			temp = ioread16(ioc9010dev->VirtureCarrierRegBaseAddr + 0x16);
			printk(KERN_NOTICE "%x Test - Carrier PCB/FPGA version = %x.\n", ioc9010dev->DEVICE_ID, temp);
		}
		else if(ioc9010dev->DEVICE_ID == 0x7002)				/* get the serial number of the carrier */
		{
			serialNum = inw(ioc9010dev->CarrierAddr + 0x04);
			while (devtemp != NULL)
			{
				if (devtemp->devnum == serialNum)
				{ 
    				printk(KERN_NOTICE "%x Test - Duplicated carrier ID %d found (Not registered). Please check carrier ID jumpers.\n", devtemp->DEVICE_ID, serialNum);
					devlist = ioc9010dev->next;
					kfree(ioc9010dev);							// free the current 
					return(-1);
				}
				devtemp = devtemp->next;
			} 
			dtMajorMinor = MKDEV(ioc_major, ioc_minor + serialNum);
    			printk(KERN_NOTICE "%x Test - Carrier ID=%d.\n", ioc9010dev->DEVICE_ID, serialNum);
			temp = inw(ioc9010dev->CarrierAddr + 0x14);
			printk(KERN_NOTICE "%x Test - Carrier card model = %x.\n", ioc9010dev->DEVICE_ID, temp);

			temp = inw(ioc9010dev->CarrierAddr + 0x16);
			printk(KERN_NOTICE "%x Test - Carrier PCB/FPGA version = %x.\n", ioc9010dev->DEVICE_ID, temp);
		}
		else if(ioc9010dev->DEVICE_ID == 0x6335)				/* get the serial number of the carrier */
		{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))										//Before Fedora 15. May need earlier version here?
			serialNum = inw(ioc9010dev->CarrierAddr + 0x04);
#else
			serialNum = inb(ioc9010dev->CarrierAddr + 0x04);
#endif    	      
			while (devtemp != NULL)
			{
				if (devtemp->devnum == serialNum)
				{ 
    				printk(KERN_NOTICE "%x Test - Duplicated carrier ID %d found (Not registered). Please check carrier ID jumpers.\n", devtemp->DEVICE_ID, serialNum);
					devlist = ioc9010dev->next;
					kfree(ioc9010dev);							// free the current 
					return(-1);
				}
				devtemp = devtemp->next;
			} 
			dtMajorMinor = MKDEV(ioc_major, ioc_minor + serialNum);
   			printk(KERN_NOTICE "%x Test - Carrier ID=%d.\n", ioc9010dev->DEVICE_ID, serialNum);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))										//Before Fedora 15. May need earlier version here?
			temp = inw(ioc9010dev->CarrierAddr + 0x14);
			printk(KERN_NOTICE "%x Test - Carrier card model = %x.\n", ioc9010dev->DEVICE_ID, temp);

			temp = inw(ioc9010dev->CarrierAddr + 0x16);
			printk(KERN_NOTICE "%x Test - Carrier PCB/FPGA version = %x.\n", ioc9010dev->DEVICE_ID, temp);
#else
			temp = (inb(ioc9010dev->CarrierAddr + 0x15) << 8) + inb(ioc9010dev->CarrierAddr + 0x14);
			printk(KERN_NOTICE "%x Test - Carrier card model = %x.\n", ioc9010dev->DEVICE_ID, temp);
			temp = (inb(ioc9010dev->CarrierAddr + 0x17) << 8) + inb(ioc9010dev->CarrierAddr + 0x16);
			printk(KERN_NOTICE "%x Test - Carrier PCB/FPGA version = %x.\n", ioc9010dev->DEVICE_ID, temp);
#endif    	      
		}
		else												/* for other single carrier */
			dtMajorMinor = MKDEV(ioc_major, ioc_minor);

		devtemp = 0;
		ioc9010dev->devnum = serialNum;

		/* add device */
		err = cdev_add(&ioc9010dev->cdev,dtMajorMinor,1);
		if (err)
			printk(KERN_NOTICE "%x Test - Error %d Registering Device\n", ioc9010dev->DEVICE_ID, err);
		else
			printk(KERN_NOTICE "%x Test - Device Registered OK\n", ioc9010dev->DEVICE_ID);


    	/* Map the Base Address for the IP Cards */
    	/* ------------------------------------- */	
    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_3, &iBaseAddressVariable);
    	printk(KERN_NOTICE "%x Test - Base Add3 - 0x%8X \n", ioc9010dev->DEVICE_ID, iBaseAddressVariable);

    	if( !request_mem_region(iBaseAddressVariable,0xE00E00,"IOC9010 IPs"))
    	{
    		printk(KERN_NOTICE "%x Test - Memory region is also used by another driver!\n", ioc9010dev->DEVICE_ID);
    	}
    	else
    	{
    		printk(KERN_NOTICE "%x Test - Memory for IPs Mapped !\n", ioc9010dev->DEVICE_ID);
			ioc9010dev->bar3State = 1;														//BAR3 mapped successfully
    	}

		ioc9010dev->IOCMemoryAddr = iBaseAddressVariable;			//save it for releasing
    	ioc9010dev->VirtureMemBaseAddr = ioremap(iBaseAddressVariable,0xE00E00);	

		if ((ioc9010dev->DEVICE_ID != 0x7003) && (ioc9010dev->DEVICE_ID != 0x7002))
		{
    		/* Map the Base Address for the PLX Local Configuration Registers I/O region*/
    		/********************** Added by Jim on 26-03-2009 **********************/
    		pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &iCarrierBoardBaseAddress);
	
    		/* These are I/O, hence LSB is only se 0x    E000 t to indicate an I/O Region */
    		iCarrierBoardBaseAddress--;				//bit0 = 1 means I/O space, should mask this bit
    		printk(KERN_NOTICE "%x Test - Base Add1 - 0x%8X \n",ioc9010dev->DEVICE_ID, iCarrierBoardBaseAddress);
    		if( !request_region(iCarrierBoardBaseAddress,80,"Hytec PLX Local"))
    		{ 
    			printk(KERN_NOTICE "%x Test - PLX Local region is also used by another driver!\n",ioc9010dev->DEVICE_ID);
    		}
    		else
    		{
    			printk(KERN_NOTICE "%x Test - PLX Local Config Registers Mapped !\n",ioc9010dev->DEVICE_ID);
				ioc9010dev->bar1State = 1;														//BAR1 mapped successfully
    		}
		
			ioc9010dev->PLXLocalConfigBAR = iCarrierBoardBaseAddress;
		}
		
	    /* Scan for IP Slots on a PCIe 6335 - Step Through the Possible ID Address Spaces */
		if(ioc9010dev->DEVICE_ID == 0x6335)
		{
			iCardCount = 0;
	    	for (IPOffset=0; IPOffset<2; IPOffset++)
	    	{
				for(i=0;i<14;i++)
				{
	    			sbuf[i] = ioread16(ioc9010dev->VirtureMemBaseAddr + IP_IO_BASE_ADDR + IP_A_ID_BASE_ADDR + 
						IPOffset*(IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) + i*2); 
/*********************************************************************************************************
if((i<8) && (IPOffset<2))
{
	    serialNum = ioread16(ioc9010dev->VirtureMemBaseAddr + IP_IO_BASE_ADDR +  
						IPOffset*(IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) + i*2); 
	    printk(KERN_NOTICE "6335 Test - reg %d value before write - %4X! \n",i, serialNum);

        iowrite16(0x55AA, ioc9010dev->VirtureMemBaseAddr  + IP_IO_BASE_ADDR + 
						IPOffset*(IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) + i*2); 

	    serialNum = ioread16(ioc9010dev->VirtureMemBaseAddr + IP_IO_BASE_ADDR +  
						IPOffset*(IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) + i*2); 
	    printk(KERN_NOTICE "6335 Test -------- %x written to %d now value - %x!\n",0x55AA, i, serialNum);
}
*********************************************************************************************************/
				}
	    		printk(KERN_NOTICE "6335 Test - Slot %c ID Found - %4X!\n",'A'+IPOffset, sbuf[ID_MODEL_NUMBER/2]);
	    		if (sbuf[ID_MODEL_NUMBER/2] != 0)
	    		{
	    			ioc9010dev->IPCard[iCardCount] = sbuf[ID_MODEL_NUMBER/2];
	    		}
	    		else
	    		{
	    			ioc9010dev->IPCard[iCardCount] = 0;
	    		}
	    		iCardCount++;
	    	}		
		}

	    /* Scan for IP Slots on 7002 uTCA - Step Through the Possible ID Address Spaces */
		if(ioc9010dev->DEVICE_ID == 0x7002)
		{
	    	for (	IPOffset  = IP_A_ID_BASE_ADDR; 
	    		IPOffset < IP_B_ID_BASE_ADDR; 
	    		IPOffset += (IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR)	)
	    	{
	    		DataRead = ioread16(ioc9010dev->VirtureMemBaseAddr + IP_IO_BASE_ADDR + IPOffset + ID_MODEL_NUMBER); 
	    		
	    		printk(KERN_NOTICE "7002 Test - Slot %c ID Found - %4X !\n",'A'+iCardCount, DataRead);

	    		if (DataRead != 0)
	    		{
	    			ioc9010dev->IPCard[iCardCount] = DataRead;
	    		}
	    		else
	    		{
	    			ioc9010dev->IPCard[iCardCount] = 0;
	    		}
	    		iCardCount++;
	    	}
		
		}

	    /* Scan for IP Slots on 7003 uTCA - Step Through the Possible ID Address Spaces */
		if(ioc9010dev->DEVICE_ID == 0x7003)
		{
			iCardCount = 0;
	    	for (IPOffset=0; IPOffset<3; IPOffset++)
	    	{
				for(i=0;i<14;i++)
	    			sbuf[i] = ioread16(ioc9010dev->VirtureMemBaseAddr + IP_IO_BASE_ADDR + IP_A_ID_BASE_ADDR + 
						IPOffset*(IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) + i*2); 
	    		
	    		printk(KERN_NOTICE "7003 Test - Slot %c ID Found - %4X!\n",'A'+IPOffset, sbuf[ID_MODEL_NUMBER/2]);
	    		if (sbuf[ID_MODEL_NUMBER/2] != 0)
	    		{
	    			ioc9010dev->IPCard[iCardCount] = sbuf[ID_MODEL_NUMBER/2];
	    		}
	    		else
	    		{
	    			ioc9010dev->IPCard[iCardCount] = 0;
	    		}
	    		iCardCount++;
	    	}
		}

		if(ioc9010dev->DEVICE_ID == 0x9010)
		{
	    	/* Overwrite LCD Display written by Xilinx - so the user can tell the Kernel's Running */
	    	Clear_LCD(ioc9010dev);
	    	Write_LCD(ioc9010dev, (LCD_CENTRE_ON | LCD_LINE_1), "Hytec Electronics Ltd");
	    	Write_LCD(ioc9010dev, (LCD_CENTRE_ON | LCD_LINE_2), "IOC Running - OK");
		
	    	/* Tell the World - we are about to Check what IP Cards we have fitted */
	    	msleep(2);
	    	Clear_LCD(ioc9010dev);
	    	Write_LCD(ioc9010dev, (LCD_CENTRE_ON | LCD_LINE_1), "Hytec Electronics Ltd");
	    	Write_LCD(ioc9010dev, (LCD_CENTRE_ON | LCD_LINE_2), "IP Slots - Scanning");


	    	/* Scan for IP Slots - Step Through the Possible ID Address Spaces */
	    	for (	IPOffset  = IP_A_ID_BASE_ADDR; 
	    		IPOffset <= IP_F_ID_BASE_ADDR; 
	    		IPOffset += (IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR)	)
	    	{
	    		DataRead = ioread16(ioc9010dev->VirtureMemBaseAddr + IP_IO_BASE_ADDR + IPOffset + ID_MODEL_NUMBER); 
	    		
	    		printk(KERN_NOTICE "9010 Test - Slot %c ID Found - %4X !\n",'A'+iCardCount, DataRead);

	    		if (DataRead != 0)
	    		{
	    			ioc9010dev->IPCard[iCardCount] = DataRead;
		 		}
	    		else
	    		{
	    			ioc9010dev->IPCard[iCardCount] = 0;
	    		}

	    		/* Debug Display Carrier Board CSR to check for timeout error 
	            DataRead = inw(ioc9010dev->CarrierAddr + REG_CSR);
	    		printk(KERN_NOTICE "9010 Test - CSR = 0x%4X \n",DataRead); */

	    		iCardCount++;
	    	}
		
	    	/* Make Up Some Strings containing the IP Card Inforamtion */
	    	msleep(2);
	    	Clear_LCD(ioc9010dev);
	    	sprintf(String0,"A - %4X  B - %4X  C - %4X",ioc9010dev->IPCard[0],ioc9010dev->IPCard[1],ioc9010dev->IPCard[2]);
	    	sprintf(String1,"D - %4X  E - %4X  F - %4X",ioc9010dev->IPCard[3],ioc9010dev->IPCard[4],ioc9010dev->IPCard[5]);

	    	/* Display IP Card IDs on the LCD */
	    	Write_LCD(ioc9010dev, (LCD_CENTRE_ON | LCD_LINE_1), String0);
	    	Write_LCD(ioc9010dev, (LCD_CENTRE_ON | LCD_LINE_2), String1);

	    	/* Tell the World - We Have Finished Self Test */
	    	msleep(2);

	    	Clear_LCD(ioc9010dev);
	    	Write_LCD(ioc9010dev, (LCD_CENTRE_ON | LCD_LINE_1), "Hytec Electronics Ltd");
	    	Write_LCD(ioc9010dev, (LCD_CENTRE_ON | LCD_LINE_2), "Self Test - OK");
		}
    	
		/* Check the PCI Carrier Interrupts */
    	/* -------------------------------- */
        pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &ioc9010dev->CarrierIRQ);
	
    	// Update Log about it all went
    	printk(KERN_NOTICE "%x Test - IRQ Pin - %d \n", ioc9010dev->DEVICE_ID, ioc9010dev->CarrierIRQ);

		ioc9010dev->CarrierIRQ = dev->irq;
		ioc9010dev->ppciDev = dev;						/* store pci device reference */
    	printk(KERN_NOTICE "%x Test - IRQ level- %d\n", ioc9010dev->DEVICE_ID, ioc9010dev->CarrierIRQ);

        // Request the IRQ and link it to the Interrupt Handler
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23))
    	result = request_irq(ioc9010dev->CarrierIRQ, InterruptHandler, SA_INTERRUPT | SA_SHIRQ, "Hytec Interrupt Handler", ioc9010dev);
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,43))			//Before Fedora 15. May need earlier version here?
        result = request_irq(ioc9010dev->CarrierIRQ, InterruptHandler, IRQF_TRIGGER_RISING| IRQF_TRIGGER_HIGH | IRQF_SHARED, "Hytec Interrupt Handler", ioc9010dev);
#else
        result = request_irq(ioc9010dev->CarrierIRQ, InterruptHandler, IRQF_SHARED, "Hytec Interrupt Handler", ioc9010dev);        
#endif
        if (result)
        {
            printk(KERN_NOTICE "%x Test - Int Handler can NOT assigned IRQ %d - %d!\n", ioc9010dev->DEVICE_ID, ioc9010dev->CarrierIRQ, result);
    	}
        else
        {
            printk(KERN_NOTICE "%x Test - Interrupt Handler assigned to IRQ %d !\n", ioc9010dev->DEVICE_ID, ioc9010dev->CarrierIRQ);
			ioc9010dev->irqState = 1;														//irq assigned successfully
    	}

		break;

    case 0x5331:
    case 0x5332:
		dtMajorMinor = MKDEV(ioc_major, ioc_minor);

		/* add device */
		err = cdev_add(&ioc9010dev->cdev,dtMajorMinor,1);
		if (err)
			printk(KERN_NOTICE "HYTEC 9010/5331/5332/6335/7002/7003 Probe - Error %d Registering Device\n",err);
		else
			printk(KERN_NOTICE "HYTEC 9010/5331/5332/6335/7002/7003 Probe - Device Registered OK\n");

    	/* Map the Base Address for the 9010's PCI Carrier Card Registers */
    	/* -------------------------------------------------------------- */
    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_3, &iCarrierBoardBaseAddress);
	
    	/* These are I/O, hence LSB is only se 0x    E000 t to indicate an I/O Region */
    	iCarrierBoardBaseAddress--;
	
    	/* Update Log about it all went */
    	printk(KERN_NOTICE "%x Test - Base Add3 - 0x%8X \n",ioc9010dev->DEVICE_ID, iCarrierBoardBaseAddress);
    	if( !request_region(iCarrierBoardBaseAddress,64,"Hytec Carrier"))
    	{ 
    		printk(KERN_NOTICE "%x Test - Carrier NOT Mapped !\n",ioc9010dev->DEVICE_ID);
    	}
    	else
    	{
    		printk(KERN_NOTICE "%x Test - Carrier Mapped !\n",ioc9010dev->DEVICE_ID);
			ioc9010dev->bar3State = 1;														//BAR3 mapped successfully
    	}

    	/* Dispite Copius Notes about Mapping the only way to get it to work is NOT to ? */
    	ioc9010dev->CarrierAddr =iCarrierBoardBaseAddress;


    	/* Map the Base Address for the 5331 Card Local Configuration Registers I/O region*/
    	/********************** Added by Jim on 26-03-2009 **********************/
    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &iCarrierBoardBaseAddress);
	
    	/* These are I/O, hence LSB is only se 0x    E000 t to indicate an I/O Region */
    	iCarrierBoardBaseAddress--;
	
    	/* Update Log about it all went */
    	printk(KERN_NOTICE "%x Test - Base Add1 - 0x%8X \n",ioc9010dev->DEVICE_ID, iCarrierBoardBaseAddress);
    	if( !request_region(iCarrierBoardBaseAddress,80,"Hytec PLX Local"))
    	{ 
    		printk(KERN_NOTICE "%x Test - Local Config Registers NOT Mapped !\n",ioc9010dev->DEVICE_ID);
    	}
    	else
    	{
    		printk(KERN_NOTICE "%x Test - Local Config Registers Mapped !\n",ioc9010dev->DEVICE_ID);
			ioc9010dev->bar1State = 1;														//BAR1 mapped successfully
    	}
		
		ioc9010dev->PLXLocalConfigBAR = iCarrierBoardBaseAddress;


    	/* Check the PCI Carrier Interrupts */
    	/* -------------------------------- */
        pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &ioc9010dev->CarrierIRQ);
	
    	// Update Log about it all went
    	printk(KERN_NOTICE "%x Test - IRQ Pin - %d \n",ioc9010dev->DEVICE_ID, ioc9010dev->CarrierIRQ);
	
		ioc9010dev->CarrierIRQ = dev->irq;
		ioc9010dev->ppciDev = dev;						/* store pci device reference */
    	printk(KERN_NOTICE "%x Test - IRQ level- %d\n", ioc9010dev->DEVICE_ID, ioc9010dev->CarrierIRQ);
	
        // Request the IRQ and link it to the Interrupt Handler
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23))
    	result = request_irq(ioc9010dev->CarrierIRQ, InterruptHandler, SA_INTERRUPT | SA_SHIRQ, "Hytec Interrupt Handler", ioc9010dev);
#else
        result = request_irq(ioc9010dev->CarrierIRQ, InterruptHandler, IRQF_TRIGGER_RISING| IRQF_TRIGGER_HIGH | IRQF_SHARED, "Hytec Interrupt Handler", ioc9010dev);
#endif
        if (result)
        {
            printk(KERN_NOTICE "%x Test - Int Handler can NOT assigned IRQ %d !\n",ioc9010dev->DEVICE_ID, ioc9010dev->CarrierIRQ);
    	}
        else
        {
            printk(KERN_NOTICE "%x Test - Interrupt Handler assigned to IRQ %d !\n",ioc9010dev->DEVICE_ID, ioc9010dev->CarrierIRQ);
			ioc9010dev->irqState = 1;														//irq assigned successfully
    	}

        break;

    case 0x0040:
		dtMajorMinor = MKDEV(ioc_major, ioc_minor);

		/* add device */
		err = cdev_add(&ioc9010dev->cdev,dtMajorMinor,1);
		if (err)
			printk(KERN_NOTICE "HYTEC 9010/5331/5332/6335/7002/7003 Probe - Error %d Registering Device\n",err);
		else
			printk(KERN_NOTICE "HYTEC 9010/5331/5332/6335/7002/7003 Probe - Device Registered OK\n");

    	/* Log the Base Address from the PCI Configuration Registers */
    	/* --------------------------------------------------------- */
    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &iBaseAddressVariable);
    	printk(KERN_NOTICE "9010 Test - Base Add0 - 0x%8X \n", iBaseAddressVariable);
    
    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &iBaseAddressVariable);
    	printk(KERN_NOTICE "9010 Test - Base Add1 - 0x%8X \n", iBaseAddressVariable);
	
    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_2, &iBaseAddressVariable);
    	printk(KERN_NOTICE "9010 Test - Base Add2 - 0x%8X \n", iBaseAddressVariable);

    	ioc9010dev->VirtureMemBaseAddr = ioremap(iBaseAddressVariable, 0x4300);	

	
    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_3, &iBaseAddressVariable);
    	printk(KERN_NOTICE "9010 Test - Base Add3 - 0x%8X \n", iBaseAddressVariable);

    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_4, &iBaseAddressVariable);
    	printk(KERN_NOTICE "9010 Test - Base Add4 - 0x%8X \n", iBaseAddressVariable);

    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_5, &iBaseAddressVariable);
    	printk(KERN_NOTICE "9010 Test - Base Add5 - 0x%8X \n", iBaseAddressVariable);

    	/* Scan for IP Slots - Step Through the Possible ID Address Spaces */
    	for (	IPOffset = 0; 
        		IPOffset < 1; 
        		IPOffset ++   )
        {
    		printk(KERN_NOTICE "9010 Test - Address - %4p !\n",
                                   (ioc9010dev->VirtureMemBaseAddr + 0x1100 + (IPOffset * 0x1000) + ID_MODEL_NUMBER));

            DataRead = ioread16(ioc9010dev->VirtureMemBaseAddr + 0x1100 + (IPOffset * 0x1000) + ID_MODEL_NUMBER); 
		
    		printk(KERN_NOTICE "9010 Test - Slot %c ID Found - %4x !\n",'A'+IPOffset, DataRead);
    	}


        /* Set IP Cards to 32MHz */
        iowrite8(0x03, ioc9010dev->VirtureMemBaseAddr + 0x0500); 
        

    	/* Scan for IP Slots - Step Through the Possible ID Address Spaces */
    	for (	IPOffset = 0; 
        		IPOffset < 1; 
        		IPOffset ++   )
        {
    		printk(KERN_NOTICE "9010 Test - Address - %4p !\n",
                                   (ioc9010dev->VirtureMemBaseAddr + 0x1100 + (IPOffset * 0x1000) + ID_MODEL_NUMBER));

            DataRead = ioread16(ioc9010dev->VirtureMemBaseAddr + 0x1100 + (IPOffset * 0x1000) + ID_MODEL_NUMBER); 
    		printk(KERN_NOTICE "9010 Test - Slot %c ID Found - %4x !\n",'A'+IPOffset, DataRead);
    	}

        break;        	
    
    default:
		dtMajorMinor = MKDEV(ioc_major, ioc_minor);

		/* add device */
		err = cdev_add(&ioc9010dev->cdev,dtMajorMinor,1);
		if (err)
			printk(KERN_NOTICE "HYTEC 9010/5331/5332/6335/7002/7003 Probe - Error %d Registering Device\n",err);
		else
			printk(KERN_NOTICE "HYTEC 9010/5331/5332/6335/7002/7003 Probe - Device Registered OK\n");

    	/* Unknown Card Simply... */
    	/* Display the Base Address from the PCI Configuration Registers */
    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &iBaseAddressVariable);
    	printk(KERN_NOTICE "9010 Test - Base Add0 - 0x%8X \n", iBaseAddressVariable);
    
    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &iBaseAddressVariable);
    	printk(KERN_NOTICE "9010 Test - Base Add1 - 0x%8X \n", iBaseAddressVariable);
	
    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_2, &iBaseAddressVariable);
    	printk(KERN_NOTICE "9010 Test - Base Add2 - 0x%8X \n", iBaseAddressVariable);

    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_3, &iBaseAddressVariable);
    	printk(KERN_NOTICE "9010 Test - Base Add3 - 0x%8X \n", iBaseAddressVariable);

    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_4, &iBaseAddressVariable);
    	printk(KERN_NOTICE "9010 Test - Base Add4 - 0x%8X \n", iBaseAddressVariable);

    	pci_read_config_dword(dev, PCI_BASE_ADDRESS_5, &iBaseAddressVariable);
    	printk(KERN_NOTICE "9010 Test - Base Add5 - 0x%8X \n", iBaseAddressVariable);

        break;        	
    
    } /* End of Device ID Switch Statement */     
	
    return 0;
}

/**********************************************************

    NAME:           remove( )

    Description:

    Function - 	clean up any allocated resources.


    Parameters:

    Inputs:
    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 23-JAN-2006     DAN     1.1         Intial Version.
 22-NOV-2010     JSC     1.2         Added removing all devices
									 Fixed bug that didn't delete cdev before
***********************************************************/
static void remove(struct pci_dev *dev)
{
	struct IOC9010_dev* ioc9010dev = devlist; 
	struct IOC9010_dev* devtemp; 

	while (ioc9010dev != NULL)
	{
		if(ioc9010dev->ppciDev == dev)
		{
		    switch (ioc9010dev->DEVICE_ID)
		    {
		    case 0x9010:
		    case 0x6335:
		    case 0x7002:
		    case 0x7003:
		        /* Release the IRQ */
				if (ioc9010dev->irqState == 1)
				{
					free_irq(ioc9010dev->CarrierIRQ, ioc9010dev);
					printk(KERN_NOTICE "%x Test - Exit: Interrupt resource released. \n", ioc9010dev->DEVICE_ID);
				}
	    			
				/* Tidy Up for Stuff Associated with Carrier Board */
				/* ----------------------------------------------- */
				/* Un Map Memory */ 
				if(ioc9010dev->bar3State == 1)
				{
					iounmap(ioc9010dev->VirtureMemBaseAddr);	
					release_mem_region(ioc9010dev->IOCMemoryAddr,0xE00E00);
					printk(KERN_NOTICE "%x Test - Exit: Memory unmapped. \n", ioc9010dev->DEVICE_ID);
				}

				if (ioc9010dev->bar1State == 1)
				{
					release_region(ioc9010dev->PLXLocalConfigBAR,80);
					printk(KERN_NOTICE "%x Test - Exit: PLX local I/O unmapped. \n", ioc9010dev->DEVICE_ID);
				}

				if (ioc9010dev->bar2State == 1)
				{
					if (ioc9010dev->DEVICE_ID == 0x7003)
					{
						iounmap(ioc9010dev->VirtureCarrierRegBaseAddr);
						release_mem_region(ioc9010dev->CarrierAddr, 64);
					}else
						release_region(ioc9010dev->CarrierAddr, 64);
					printk(KERN_NOTICE "%x Test - Exit: Carrier I/O unmapped. \n", ioc9010dev->DEVICE_ID);
				}
	
/*
				if(ioc9010dev->DEVICE_ID == 0x6335)
				{
					iounmap(addrPCI);	
					release_mem_region(ioc9010dev->PCIexpressBAR,0xFFFF);			
				}
*/
		        break;
	
		    case 0x5331:
		    case 0x5332:
				/* Tidy Up for Stuff Associated with Carrier Board */
				/* ----------------------------------------------- */
				if(ioc9010dev->bar3State == 1) release_region(ioc9010dev->CarrierAddr,64);
				if(ioc9010dev->bar1State == 1) release_region(ioc9010dev->PLXLocalConfigBAR,80);
	    
		        /* Release the IRQ */
		        if(ioc9010dev->irqState == 1) free_irq(ioc9010dev->CarrierIRQ, NULL);
		        break;
	
		    case 0x0040:
		    default:
		    	/* Release Memory Region - Is this needed ? */
		    	/* release_mem_region((unsigned int)(ioc9010dev->VirtureMemBaseAddr),0x4300); */
	    
		        /* Release the IRQ */
		        /* free_irq(ioc9010dev->CarrierIRQ, NULL); */
		        break;
		    }

			printk(KERN_NOTICE "%x Test - The remove is called \n", ioc9010dev->DEVICE_ID);

			pci_disable_device(dev);
			ioc9010dev->ppciDev = NULL;
			devtemp = ioc9010dev->next;						/* retrieve next */
			cdev_del(&ioc9010dev->cdev);						/* delete char device */
			kfree(ioc9010dev);							/* free the current */
			ioc9010dev = devtemp;							/* try the next one */
		}else
			ioc9010dev = ioc9010dev->next;
	}
}


static struct pci_driver pci_driver = {
	.name = "pci_hytec",
	.id_table = ids,
	.probe = probe,
	.remove = remove,
};


/**********************************************************

    NAME:           pci_hytec_init( )

    Description:

    Function - 


    Parameters:

    Inputs:
    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham
					Jim Chen

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 23-JAN-2006     DAN     1.1         Intial Version.
 22-NON-2010     JSC     1.2         Moved number registration to here.

***********************************************************/
static int __init pci_hytec_init(void)
{
	int result;
	printk(KERN_NOTICE "Entry of Hytec Linux Device Driver - pci inited \n");

	/* Allocate device number */
	/* If you want to set your own.....*/
	ioc_major = 0;						/* No, force it to use dynamic numbering in complaint to the main stream. JC 06/12/2010 */

	if (ioc_major) 
	{
		dtMajorMinor = MKDEV(ioc_major, ioc_minor);
		result = register_chrdev_region(dtMajorMinor, MAXDEVNUM, "IOC9010");
	}
	/* Dynamically Allocate one instead...*/
	else 
	{    
		result = alloc_chrdev_region(&dtMajorMinor, ioc_minor, MAXDEVNUM,"IOC9010");
		ioc_major = MAJOR(dtMajorMinor);
	}
	if (result < 0) 
	{
		printk(KERN_WARNING "HYTEC 9010/5331/5332/6335/7002/7003 Probe - Can't get major %d\n", ioc_major);
		return result;
	}
	
	return pci_register_driver(&pci_driver);
}


/**********************************************************

    NAME:           pci_hytec_exit( )

    Description:

    Function - 


    Parameters:

    Inputs:
    Outputs:
    Return:

    Notes:

    Programmer(s):  Darrell Nineham

**********************************************************

 Revision History
 ----------------

 Date            By      Version     Descripition
 ----            --      -------     ------------
 23-JAN-2006     DAN     1.1         Intial Version.

***********************************************************/
static void __exit pci_hytec_exit(void)
{
	dev_t devno;

	pci_unregister_driver(&pci_driver);

	devno = MKDEV(ioc_major, ioc_minor);

	/* Free Memory Allocated to Kernel buffer */
	kfree(buffer);

	/* cleanup_module is never called if registering failed */
	unregister_chrdev_region(devno, MAXDEVNUM);
}



MODULE_LICENSE("GPL");

module_init(pci_hytec_init);
module_exit(pci_hytec_exit);
