#include <stdio.h>
#include "9010LinuxDriver.h"

int IOCHandle;

long drvHy8505_write(const int cardnum, const unsigned short mask, unsigned short value);


int main(void)
{
	IOCTL_BUF		ioctl_buf;
    long 			IPOffset;
	unsigned short 	IPCards[6];

	unsigned short data,count,readbuffer[21];
	
	printf("Trying to Open IOC\n");

	IOCHandle = open("/dev/IOC9010");
	if (IOCHandle == -1) printf("Error Opening Device !\n");


	/* Scan all IP Slots ID PROM */ 
	count = 0;

	for (   IPOffset  = IP_A_ID_BASE_ADDR; 
            IPOffset <= IP_F_ID_BASE_ADDR; 
            IPOffset += (IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR)	)
	{	
		ioctl_buf.lAddress = IPOffset + ID_MODEL_NUMBER;
		ioctl_buf.lLength  = 1;
		ioctl_buf.sData  = (unsigned long)(&data);

		/* Read IP Card Type from ID PROM */
		IPCards[count] = ioctl(IOCHandle, OP_GENERAL_READ, &ioctl_buf);

		printf("IP Slot %c = %4X\n", 'A'+count, IPCards[count]);

		count++;
	}

    /* Read Carrier Board Register */
	for (   IPOffset  = REG_CSR; 
            IPOffset <= REG_CONFIG_2; 
            IPOffset += 2	)
	{	
		ioctl_buf.lAddress = IPOffset;
		ioctl_buf.lLength  = 1;
		ioctl_buf.sData  = (unsigned long)(&data);

		/* Read IP Card Type from ID PROM */
		ioctl(IOCHandle, OP_CARRIER_READ_BLOCK, &ioctl_buf);

		printf("REG %2X = %4X\n", IPOffset, data);
	}

	/* Set Up the IP Cards */
	for (   count = 0; 
            count < 6; 
            count ++	)
	{
		switch (IPCards[count])
		{
		case 0x8401:

			/* 8401 Basic Analog Input Setup */
			readbuffer[0] = 0x8000;	/* Register 0 - Initial Output Value */
			readbuffer[1] = 0x0000;	/* Register 2 */
			readbuffer[2] = 0x0000; /* Register 4 */
			readbuffer[3] = 0x0006; /* Register 6 */

			ioctl_buf.lAddress = IP_A_IO_BASE_ADDR + ((IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) * count);
			ioctl_buf.lLength  = 4;
			ioctl_buf.sData    = (unsigned long)(readbuffer);
			ioctl(IOCHandle, OP_GENERAL_WRITE_BLOCK, &ioctl_buf); 
			break;

		case 0x8402:

			/* 8402 Basic Analog Output Setup */
			readbuffer[0] = 0x8000;	/* Register 0 - Initial Output Value */

			ioctl_buf.lAddress = IP_A_IO_BASE_ADDR + ((IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) * count);
			ioctl_buf.lLength  = 1;
			ioctl_buf.sData    = (unsigned long)(readbuffer);
			ioctl(IOCHandle, OP_GENERAL_WRITE_BLOCK, &ioctl_buf); 
			break;

		case 0x8505:

			/* 8505 Basic Digital Output Setup */
			readbuffer[0] = 0xEEEE;	/* Register 0 - Initial Output Value */
			readbuffer[1] = 0x0C01;	/* Register 2 */
			readbuffer[2] = 0x0078; /* Register 4 */
			readbuffer[3] = 0x0000; /* Register 6 */
			readbuffer[4] = 0x0000; /* Register 8 */

			/* 8505 Basic Digital Input Setup */
			/*
			readbuffer[0] = 0x0001;	// Register 2
			readbuffer[1] = 0x0078; // Register 4
			readbuffer[2] = 0x0000; // Register 6
			readbuffer[3] = 0x00FF; // Register 8
			*/
			ioctl_buf.lAddress = IP_A_IO_BASE_ADDR + ((IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) * count);
			ioctl_buf.lLength  = 5;
			ioctl_buf.sData    = (unsigned long)(readbuffer);
			ioctl(IOCHandle, OP_GENERAL_WRITE_BLOCK, &ioctl_buf); 

			break;

		default:
			break;
		}

	}


	/* Write Digital Out */
	/* 
	data = 0xCCCC;
	ioctl_buf.lAddress = IP_A_IO_BASE_ADDR + ((IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) * count);
	ioctl_buf.lLength  = 1;
	ioctl_buf.sData    = (unsigned long)(&data);
	ioctl(IOCHandle, OP_GENERAL_WRITE_BLOCK, &ioctl_buf);
	*/

	/* ---------- Read I/O Registers from IP Card D ------------ */
	/*	
	ioctl_buf.lAddress = IP_D_IO_BASE_ADDR;
	ioctl_buf.lLength  = 7;
	ioctl_buf.sData  = (unsigned long)(readbuffer);
	ioctl(IOCHandle, OP_GENERAL_READ_BLOCK, &ioctl_buf);

	for (count = 0; count < 7; count++)
    {   
        printf("ID Reg %2d - %04X\n", 	count*2, 
										readbuffer[count]);
    }
	*/

	/* ---------- Test using IOC Kernal Function Read ------------ */
	/*

    char* CarrierBoardRegisters[] = {
    "CSR      ",
    "CONFIG   ",
    "DISP_CONT",
    "DISP_DATA",
    "INTS_LO  ",
    "INTS_HI  ",
    "MASK_LO  ",
    "MASK_HI  ",
    "IP_CLK   ",
    "FAN_1_2  ",
    "FAN_3_4  ",
    "FAN_5_6  ",
    "FAN_CONT ",
    "TEMP_FLAG",
    "CONFIG_2 "
    };

	read(IOCHandle, readbuffer, 42);

	for (count = 0; count < 15; count++)
    {   
        printf("Carrier Reg %s - %02X = %04X\n", CarrierBoardRegisters[count], count*2, readbuffer[count]);
    }

	for (count = 0; count < 6; count++)
    {   
        printf("IP %c = %X\n", 'A'+count, readbuffer[count+15]);
    }
	*/

	/* ------------- Read the ID PROM of IP A ----------------- */
	/*	
	ioctl_buf.lAddress = IP_A_ID_BASE_ADDR + ID_MODEL_NUMBER;
	ioctl_buf.lLength  = 1;
	ioctl_buf.sData  = (unsigned long)(&data);

	printf("IP Slot A = %X\n", ioctl(IOCHandle, OP_GENERAL_READ, &ioctl_buf));


	ioctl_buf.lAddress = IP_A_ID_BASE_ADDR;
	ioctl_buf.lLength  = 21;
	ioctl_buf.sData  = (unsigned long)(readbuffer);

	ioctl(IOCHandle, OP_GENERAL_READ_BLOCK, &ioctl_buf);

	for (count = 0; count < 15; count++)
    {   
        printf("ID Reg %2d - '%c%c'\t%02X\n", 	count*2, 
												(readbuffer[count] >> 8),
												(readbuffer[count] & 0xFF), 
												readbuffer[count]);
    }
	*/

	drvHy8505_write(2, 0xFF, 0x2211);

	close(IOCHandle);

	return 0;
}


long drvHy8505_write(const int cardnum, const unsigned short mask, unsigned short value)
{
	IOCTL_BUF		ioctl_buf;
	unsigned short data = value;

	ioctl_buf.lAddress = IP_A_IO_BASE_ADDR + ((IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) * cardnum);
	ioctl_buf.lLength  = 1;
	ioctl_buf.sData    = (unsigned long)(&data);
	ioctl(IOCHandle, OP_GENERAL_WRITE_BLOCK, &ioctl_buf);
}
