/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770	*/
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566	*/
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/
/*                                                                              */
/*				IOC9010 PROJECT	--- Hytec IP Card API Functions 				*/
/*							for Windows and Linux Platforms	    				*/
/*                                IP Card Type: 85156                           */
/********************************************************************************/
/*  Source file name      :-      Hy85156.h										*/
/*  Definition header file:		  HYTECIPAPI.h									*/
/*  Platform			  :		  Linux, Windows XP, Windows 2000               */
/*  Initial creation date :-      11-AUG-2006                                   */
/*                                                                              */
/*  Original Developers   :-      Jim S Chen                                    */
/********************************************************************************/
/*                                                                              */
/*  Description :-                                                              */
/*                                                                              */
/*  This header file is designed for Hytec API function library of 8515 and		*/
/*	8516 IP card. It defines the class prototype, error code, masks and 		*/
/*	registers of 8515/8516.    													*/
/*																				*/
/*  The design of this interface is cross platforms, which means the code 		*/
/*	can be used in operating systems like Windows, Linux etc without changes.	*/
/*                                                                              */
/*  The language used is C++ but the final product -- the Dynamic Link library	*/
/*	(dll) or the shared library can be used by either C or C++ applications.	*/
/*																				*/
/*  (C)2006 Hytec Electronics Ltd.												*/
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Modification record :-                                                      */
/*                                                                              */
/*  Date modified |     Person    |   Description of modification | Version     */
/* ---------------|---------------|-------------------------------------------- */
/*    11/08/06    |	  Jim S Chen  |	  Initial copy				  |	0.0.1		*/
/*    26/03/09    |   JC          |   Added interrupt support     |  0.2.0      */
/*    21/10/09    |   JC          |   Added Half/full duplex      |  0.3.0      */
/*                |               |   support by overloading      |             */
/*                |               |   ComSetup and configure funs.|             */
/*                |               |   Correct baud rate           |             */
/*                |               |   setting when devided by 4.  |             */
/*    18/10/10    |   KJ          |   Added a support for than 1  |  1.0.0      */
/*    		      |   	          |   device                      |  		    */
/*                |               |                               |             */
/********************************************************************************/


//error code definition
#define NOT_85156_IP -85156010
#define Hy85156_NOT_CONFIGURED -85156011
#define PORT_INVALID_85156 -85156012
#define BAUD_RATE_INVALID -85156013
#define PARITY_INVALID -85156014
#define DATA_BITS_INVALID -85156015
#define STOP_BITS_INVALID -85156016
#define FLOW_CONTROL_INVALID -85156017
#define INT_TRIGGERLEVEL_INVALID -85156018
#define FALSE_INTERRUPT -85156019
#define RS485_DELAYERROR -85156020
#define RS485_DUPLEXERROR -85156021
#define GENERAL_PURPOSE_INVALID -85156022

//signature definition
#define SIGNATURE_8515 0x8515
#define SIGNATURE_8516 0x8516

//register address defines
#define INTERRUPT_REG_85156 0x00*2
#define INT_SOURCE1_85156 0x01*2
#define INT_SOURCE2_85156 0x02*2
#define INT_SOURCE3_85156 0x03*2
#define TIMER_CONTROL_85156 0x04*2
#define TIMER_CONTROL0_85156 0x05*2
#define TIMER_COUNTER_LS_85156 0x06*2
#define TIMER_COUNTER_MS_85156 0x07*2
#define MODE_REG_85156 0x08*2
#define DUPLEX_REG_85156 0x09*2
#define RESET_REG_85156 0x0A*2
#define SLEEP_REG_85156 0x0B*2
#define REVISION_REG_85156 0x0C*2
#define ID_REG_85156 0x0D*2
#define CSR_85156 0x0E*2

#define ENABLE_INTERRUPT_85156 0x4000
#define PORT_OFFSET 0x20

#define RHR_85156 0x00
#define THR_85156 RHR_85156
#define DLL_85156 RHR_85156
#define DLM_85156 0x01
#define IER_85156 DLM_85156
#define ISR_85156 0x02
#define FCR_85156 ISR_85156
#define LCR_85156 0x03
#define MCR_85156 0x04
#define LSR_85156 0x05
#define MSR_85156 0x06
#define SPR_85156 0x07
#define FCTR_85156 0x08
#define EFR_85156 0x09
#define TXCNT_85156 0x0A
#define TXTRG_85156 TXCNT_85156
#define RXCNT_85156 0x0B
#define RXTRG_85156 RXCNT_85156
#define RCHAR_85156 0x0C
#define XOFF1_85156 RCHAR_85156
#define XOFF2_85156 0x0D
#define XON1_85156 0x0E
#define XON2_85156 0x0F

//mask definition
#define PARITY_MASK_85156 0xC7
#define NO_PARITY_85156 0x00
#define ODD_PARITY_85156 0x08
#define EVEN_PARITY_85156 0x18

#define DATABIT_MASK_85156 0xFC
#define DATA_5_85156 0x00
#define DATA_6_85156 0x01
#define DATA_7_85156 0x02
#define DATA_8_85156 0x03

#define STOPBIT_MASK_85156 0xFB
#define STOPBIT_1_85156 0x00
#define STOPBIT_2_85156 0x04

#define ENABLE_DIVISOR_MASK_85156 0x80
#define ENABLE_FIFO_85156 0x01
#define THR_EMPTY_85156 0x20
#define TSR_EMPTY_85156 0x40

#define ENABLE_RHR_INT_85156 0x01
#define ENABLE_THR_INT_85156 0x02
#define ENABLE_RCVLINE_INT_85156 0x04
#define ENABLE_MODEM_INT_85156 0x08
#define ENABLE_XON_XOFF_INT_85156 0x20
#define ENABLE_RTS_DTR_INT_85156 0x40
#define ENABLE_CTS_DSR_INT_85156 0x80

#define FCR_RXFRESET_85156 0x02
#define FCR_TXFRESET_85156 0x04

#define ENABLE_LEVEL_TRIGGER_85156 0xC0

#define LCR_TXBREAK_85156 0x40

#define MCR_DTR_85156 0x01
#define MCR_RTS_85156 0x02
#define MCR_LOOP_85156 0x10

#define MSR_CTS_85156 0x01
#define MSR_DSR_85156 0x02
#define MSR_RI_85156 0x04
#define MSR_CD_85156 0x08

#define LSR_OVERRUN_85156 0x02
#define LSR_PARITY_85156 0x04
#define LSR_FRAME_85156 0x08
#define LSR_BREAK_85156 0x10

/******* Interrupt reasons ******/
/* Major reason */
#define INTERRUPT_REASON_MSR 0x00
#define INTERRUPT_REASON_LASER INTERRUPT_REASON_MSR                 /* This macro is purely for Sky Scanner job */
#define INTERRUPT_REASON_NONE 0x01
#define INTERRUPT_REASON_TXRDY 0x02
#define INTERRUPT_REASON_RXRDY 0x04
#define INTERRUPT_REASON_LSR 0x06
#define INTERRUPT_REASON_RXRDYTO 0x0C
#define INTERRUPT_REASON_RXRDYXONOFF 0x10
/* Detail reason */
#define INTERRUPT_REASON_MSR_CTS 0x0100
#define INTERRUPT_REASON_MSR_DSR 0x0200
#define INTERRUPT_REASON_MSR_RI 0x0400
#define INTERRUPT_REASON_MSR_CD 0x0800
#define INTERRUPT_REASON_LASER_READY INTERRUPT_REASON_MSR_CD        /* This macro is purely for Sky Scanner job */

#define INTERRUPT_REASON_LSR_RX_OVERRUN 0x0200
#define INTERRUPT_REASON_LSR_RX_PARITY_ERROR 0x0400
#define INTERRUPT_REASON_LSR_RX_FRAMING_ERROR 0x0800
#define INTERRUPT_REASON_LSR_RX_BREAK 0x1000
/******* Interrupt reasons end******/


/* general purpose line definition */
#define GENERAL_PURPOSE_DTR_85156		0x01
#define GENERAL_PURPOSE_RTS_85156		0x02
#define GENERAL_PURPOSE_USE_DTR_85156	0x04


/* baud rate settings */
#define DIVIDED_BY_FOUR_85156			0x80
#define AUTO_RS485_ENABLE_85156			0x20

#define BAUDRATE_100_DLL_85156			0x00
#define BAUDRATE_100_DLM_85156			0x09

#define BAUDRATE_400_DLL_85156			0x00
#define BAUDRATE_400_DLM_85156			0x09

#define BAUDRATE_600_DLL_85156			0x80
#define BAUDRATE_600_DLM_85156			0x01

#define BAUDRATE_1200_DLL_85156			0xC0
#define BAUDRATE_1200_DLM_85156			0x00

#define BAUDRATE_2400_DLL_85156			0x80
#define BAUDRATE_2400_DLM_85156			0x00

#define BAUDRATE_4800_DLL_85156			0xC0
#define BAUDRATE_4800_DLM_85156			0x00

#define BAUDRATE_9600_DLL_85156			0x60
#define BAUDRATE_9600_DLM_85156			0x00

#define BAUDRATE_19200_DLL_85156		0x30
#define BAUDRATE_19200_DLM_85156		0x00

#define BAUDRATE_38400_DLL_85156		0x18
#define BAUDRATE_38400_DLM_85156		0x00

#define BAUDRATE_57600_DLL_85156		0x04
#define BAUDRATE_57600_DLM_85156		0x00

#define BAUDRATE_76800_DLL_85156		0x0C
#define BAUDRATE_76800_DLM_85156		0x00

#define BAUDRATE_115200_DLL_85156		0x02
#define BAUDRATE_115200_DLM_85156		0x00

#define BAUDRATE_153600_DLL_85156		0x06
#define BAUDRATE_153600_DLM_85156		0x00

#define BAUDRATE_230400_DLL_85156		0x04
#define BAUDRATE_230400_DLM_85156		0x00

#define BAUDRATE_460800_DLL_85156		0x02
#define BAUDRATE_460800_DLM_85156		0x00

#define BAUDRATE_921600_DLL_85156		0x01
#define BAUDRATE_921600_DLM_85156		0x00

#define USE_HARDWARE_FLOW_TEST_85156	0x10
#define CTS_DTR_TEST_85156				0x20
#define RTS_DTR_CTS_DSR_ENABLE_85156	0xC0
#define DTR_DSR_SELECT_85156			0x04
#define USE_SOFTWARE_FLOW_TEST_85156	0x0F
#define SOFTWARE_TRANS_FLOW_85156		0x0C
#define SOFTWARE_RCV_FLOW_85156			0x03
#define SOFTWARE_FLOW_MASK_85156				0x0F

#define FLUSH_TX_FIFO 0x04
#define FLUSH_RX_FIFO 0x02
#define COM_0_TX_FLUSH 0x0001
#define COM_0_RX_FLUSH 0x0002
#define COM_1_TX_FLUSH 0x0004
#define COM_1_RX_FLUSH 0x0008
#define COM_2_TX_FLUSH 0x0010
#define COM_2_RX_FLUSH 0x0020
#define COM_3_TX_FLUSH 0x0040
#define COM_3_RX_FLUSH 0x0080
#define COM_4_TX_FLUSH 0x0100
#define COM_4_RX_FLUSH 0x0200
#define COM_5_TX_FLUSH 0x0400
#define COM_5_RX_FLUSH 0x0800
#define COM_6_TX_FLUSH 0x1000
#define COM_6_RX_FLUSH 0x2000
#define COM_7_TX_FLUSH 0x4000
#define COM_7_RX_FLUSH 0x8000

#define FIFO_SIZE 64
#define FIFO_SIZE 64
#define MAX_PORT_NO 7



